"""This module is the interface to CRDS configuration information.
Predominantly it is used to define CRDS file cache paths and file location
functions.
"""
import os
import os.path
import re
import glob
import getpass
import configparser

# ===========================================================================

from crds.core import log, exceptions

# ===========================================================================

def _get_crds_ini_path():
    """Return the path to the CRDS rc file defining CRDS parameter settings."""
    default_rc = os.path.expanduser("~/.crds.ini")
    return os.environ.get("CRDS_INI_FILE", default_rc)

# Re-reading and re-parsing the rc file is probably too slow not to cache.
# Implement brute-force caching to avoid circular dependency with crds.utils
CRDS_INI_PARSER = None

def _get_crds_ini_parser():
    """Load and return the environment from the CRDS rc file."""
    global CRDS_INI_PARSER
    if CRDS_INI_PARSER is None:
        parser = configparser.SafeConfigParser()
        with log.warn_on_exception("Failed reading CRDS rc file"):
            ini_path = _get_crds_ini_path()
            if os.path.exists(ini_path):
                mode = os.stat(ini_path).st_mode
                if mode & 0o077:
                    raise exceptions.CrdsWebAuthenticationError("You must 'chmod 0600 $HOME/.crds.ini' to make it secret.")
                parser.read(ini_path)
                CRDS_INI_PARSER = parser
            else:
                log.verbose("No CRDS .ini file found at", repr(ini_path))
    else:
        parser = CRDS_INI_PARSER
    return parser

# Not caching this enables doing env overrides on-the-fly in os.environ,  including
# the basic machinery of ConfigItem.set() which is a non-permanent set.
def get_crds_env_str(section, varname, default):
    """Return the compsite env obtained by reading the .crds.ini file and overriding
    from os.environ.
    """
    if section:
        ini_parser = _get_crds_ini_parser()
        try:
            ini_val = ini_parser.get(section, varname)
        except configparser.NoSectionError:
            ini_val = None
        except configparser.NoOptionError:
            ini_val = None
    else:
        ini_val = None
    if ini_val in ["None", "none", "NONE"]:
        ini_val = None
    result = os.environ.get(varname, ini_val)  # os.environ overrides ini file.
    return result if result is not None else default

# ===========================================================================

def env_to_bool(varname, default=False, section=None):
    """Convert the specified environment variable `varname` into a Python bool
    defaulting to `default` if it's not defined in os.environ.
    """
    env_str = get_crds_env_str(section, varname, default)
    return env_str_to_bool(varname, env_str)

def env_str_to_bool(varname, val):
    """Convert the boolean environment value string `val` to a Python bool
    on behalf of environment variable `varname`.
    """
    if val in ["False", "false", "FALSE", "F", "f", "0", False, 0]:
        rval = False
    elif val in ["True", "true", "TRUE", "T", "t", "1", True, 1]:
        rval = True
    else:
        raise ValueError("Invalid value " +  repr(val) +
                         " for boolean env var " + repr(varname))
    return rval

def env_to_int(varname, default, section=None):
    """Convert the specified environment variable `varname` into a Python int
    defaulting to `default` if it's not defined in os.environ.
    """
    env_str = get_crds_env_str(section, varname, default)
    return env_str_to_int(varname, env_str)

def env_str_to_int(varname, val):
    """Convert environment variable decimal value `val` from `varname` to an int and return it."""
    try:
        return int(val)
    except Exception:
        raise ValueError("Invalid value for " + repr(varname) +
                         " should have a decimal integer value but is " + repr(str(val)))

# ===========================================================================

class ConfigItem:
    """CRDS environment control item base class.  These are driven by environment variables
    but conceptually could extend to a CRDS .rc file of some kind.

    >>> CFG = ConfigItem("CRDS_CFG_ITEM", "999", "Test config item", valid_values=["999", "1000"])
    >>> CFG.get()
    '999'
    >>> import os
    >>> os.environ["CRDS_CFG_ITEM"] = "1000"
    >>> CFG.get()
    '1000'
    >>> os.environ["CRDS_CFG_ITEM"] = "2"   # Don't do this!
    >>> CFG.get()
    Traceback (most recent call last):
    ...
    AssertionError: Invalid value for 'CRDS_CFG_ITEM' of '2' is not one of ['999', '1000']
    >>> os.environ["CRDS_CFG_ITEM"] = "1000"    # cleaup .set() can't do

    >>> CFG.set("999")
    '1000'
    >>> CFG.get()
    '999'
    >>> os.environ["CRDS_CFG_ITEM"]
    '999'
    """
    def __init__(self, env_var, default, comment=None, valid_values=None, lower=False, ini_section=None):
        """Defines CRDS environment item named `env_var` which has the value `default` when not specified anywhere."""
        self.env_var = env_var  # for starters,  this IS an env var,  bu conceptually it is an identifier.
        self.default = default
        self.comment = comment
        self.valid_values = valid_values
        self.lower = lower
        self.ini_section = ini_section
        self.get()

    def check_value(self, value):
        """Verify that `value` is valid for this config item."""
        if self.valid_values:
            assert value in self.valid_values, "Invalid value for " + repr(self.env_var) + " of " + repr(value)  + \
                " is not one of " + repr(self.valid_values)
        return value

    def get(self):
        """Return the value of this control item,  or the default if it is not set."""
        value = get_crds_env_str(self.ini_section, self.env_var, self.default)
        value = self._set(value)
        return value

    def set(self, value):
        """Set the value of the control item,  for the sake of this runtime session only."""
        old = self.get()
        self._set(value)
        return old

    def _set(self, value):
        """Set the value of the control item,  for the sake of this runtime session only."""
        if self.lower and isinstance(value, str):
            value = value.lower()
        self.check_value(value)
        os.environ[self.env_var] = str(value)
        return value

    def reset(self):
        """Restore this variable to it's default value,  clearing any environment setting."""
        os.environ.pop(self.env_var, None)

class StrConfigItem(ConfigItem):
    """Config item for a string value,  currently no difference from base ConfigItem.

    >>> CFG = StrConfigItem("CRDS_CFG", "something", "CRDS configuration",
    ...       valid_values=["something","else","other"])

    >>> CFG.set("foo")
    Traceback (most recent call last):
    ...
    AssertionError: Invalid value for 'CRDS_CFG' of 'foo' is not one of ['something', 'else', 'other']

    >>> CFG + "_foo"
    'something_foo'

    >>> "foo_" + CFG
    'foo_something'

    >>> CFG == 'something'
    True

    >>> 'something' == CFG
    True

    >>> CFG == 'foo'
    False

    >>> 'foo' == CFG
    False

    >>> str(CFG)
    'something'
    """

    def __str__(self):
        """Return value of string config item."""
        return str(self.get())

    def __hash__(self):
        """Use str() of StrConfigItem as hash()"""
        return hash(str(self))

    def __eq__(self, other):
        """Test string value of config item for equality with `other`."""
        return str(self.get()) == other

    def __add__(self, other):
        """Add string value of config item to following parameter `other`."""
        return str(self) + str(other)

    def __radd__(self, other):
        """Add string value of config item to preceding parameter `other`."""
        return str(other) + str(self)

class BooleanConfigItem(ConfigItem):
    """Represents a boolean environment setting for CRDS.

    >>> BOOL = BooleanConfigItem("CRDS_BOOL_ITEM", False, "Test boolean config item")
    >>> if BOOL:
    ...    print("True")
    ... else:
    ...    print("False")
    False

    >>> os.environ["CRDS_BOOL_ITEM"] = "True"
    >>> if BOOL:
    ...    print("True")
    ... else:
    ...    print("False")
    True

    >>> BOOL.set("True")  # .set() returns old value
    True
    >>> if BOOL:
    ...    print("True")
    ... else:
    ...    print("False")
    True

    >>> BOOL.set("False")
    True

    >>> if BOOL:
    ...    print("True")
    ... else:
    ...    print("False")
    False

    """
    def __init__(self, var, default, *args, **keys):
        keys = dict(keys)
        keys["valid_values"] = ["0", "1", "true", "false", "t", "f", "False", "True",
                                False, True, 0, 1]
        keys["lower"] = True
        super(BooleanConfigItem, self).__init__(var, str(default), *args, **keys)

    def get(self):
        """Return the bool value of this config item."""
        return env_str_to_bool(self.env_var, super(BooleanConfigItem, self).get())

    def set(self, val):
        """Set the bool value of this config item to `val`,  coercing to bool.  Return old value."""
        return super(BooleanConfigItem, self).set(str(env_str_to_bool(self.env_var, str(val))))

    def __nonzero__(self):
        """Support using this boolean config item be used as a conditional expression."""
        return self.get()

    __bool__ = __nonzero__


class IntConfigItem(ConfigItem):
    """Represents a boolean environment setting for CRDS.

    >>> INT = IntConfigItem(
    ...             "CRDS_INT_ITEM", '57', "Test int config item")
    >>> if INT.get():
    ...    print("True")
    ... else:
    ...    print("False")
    True

    >>> os.environ["CRDS_INT_ITEM"] = "0"
    >>> if INT.get():
    ...    print("True")
    ... else:
    ...    print("False")
    False

    >>> INT.get()
    0

    >>> INT.set("42")  # .set() returns old value
    0

    >>> INT.get()
    42
    """
    def __init__(self, var, default, *args, **keys):
        keys = dict(keys)
        keys["valid_values"] = None
        super(IntConfigItem, self).__init__(var, str(default), *args, **keys)

    def get(self):
        """Return the bool value of this config item."""
        return int(super(IntConfigItem,self).get())

    def set(self, val):
        """Set the bool value of this config item to `val`,  coercing to bool.  Return old value."""
        return super(IntConfigItem, self).set(str(int(val)))

    def __nonzero__(self):
        """Support using this boolean config item be used as
        a conditional expression."""
        return self.get() != 0

# ===========================================================================

FITS_IGNORE_MISSING_END = BooleanConfigItem("CRDS_FITS_IGNORE_MISSING_END", False,
    "When True, ignore missing END records in the FITS primary header.  Otherwise fail.")

FITS_VERIFY_CHECKSUM = BooleanConfigItem("CRDS_FITS_VERIFY_CHECKSUM", True,
    "When True, verify that FITS header CHECKSUM and DATASUM values are correct.  Otherwise fail.")

ADD_LOG_MSG_COUNTER = BooleanConfigItem(
    "CRDS_ADD_LOG_MSG_COUNTER", False, "When True, add a running counter.")
log.set_add_log_msg_count(ADD_LOG_MSG_COUNTER)

# ===========================================================================

# Runtime bad file options for end users

ALLOW_BAD_REFERENCES  = BooleanConfigItem("CRDS_ALLOW_BAD_REFERENCES", False,
    "When True, references which are designated as BAD (scientifically invalid) on the server can be used with warnings.")

ALLOW_BAD_RULES = BooleanConfigItem("CRDS_ALLOW_BAD_RULES", False,
    "When True, rules which are designated as BAD (scientifically invalid) on the server can be used with warnings.")

# ===========================================================================

# Refactoring options

PASS_INVALID_VALUES = BooleanConfigItem("PASS_INVALID_VALUES", False,
    "When True,  return invalid values detected by data model schema. These are still be errors.")

ALLOW_SCHEMA_VIOLATIONS = BooleanConfigItem("CRDS_ALLOW_SCHEMA_VIOLATIONS", False,
    "When True, don't map data model warnings onto CRDS errors.")

ALLOW_BAD_PARKEY_VALUES = BooleanConfigItem("CRDS_ALLOW_BAD_PARKEY_VALUES", False,
    "When True, turn off parkey value checking when loading rmaps.  For refactoring with bad 'legacy' references.")

ALLOW_BAD_USEAFTER = BooleanConfigItem("CRDS_ALLOW_BAD_USEAFTER", False,
    "When True,  when USEAFTER won't parse,  fake it as the primal default.")

# ============================================================================

CRDS_DATA_CHUNK_SIZE = 2**23
#   IntConfigItem("CRDS_DATA_CHUNK_SIZE", 2**23,
#   "File download transfer block size, 8M.  HTTP mode only.")

CRDS_CHECKSUM_BLOCK_SIZE = 2**23

# IntConfigItem("CRDS_CHECKSUM_BLOCK_SIZE", 2**23,
#    "Size of data read into memory at once for utils.checksum.")

# ===========================================================================

# To support testing, the default cache is configurable.  Ordinarily
# the default cache is reserved for configuration free onsite use.
# The correct approach (outside the realm of testing) for configuring
# CRDS is to use CRDS_PATH and related settings.
CRDS_DEFAULT_CACHE = os.environ.get("CRDS_DEFAULT_CACHE", "/grp/crds/cache")

def _clean_path(path):
    """Fetch `name` from the environment,  or use `default`.  Trim trailing /'s from
    the resulting path.
    """
    while path.endswith("/"):
        path = path[:-1]
    return path

def _std_cache_path(observatory, root_env, subdir):
    """Do standard interpretation of environment variables for CRDS cache component location.

    Evaluated in order,  with first definition winning:  e.g. for mappings

    CRDS_MAPPATH_SINGLE  -->   <env path for mappings>
    CRDS_MAPPATH         -->   <env path for mappings for both observatories> + / + <observatory>
    CRDS_PATH_SINGLE     -->   <env path for overall cache for one observatory> + "/mappings/"
    CRDS_PATH            -->   <env path for overall cache for all observatories> + "/mappings/" + <observatory>

    XXXX `observatory` should accept pseudo-observatory "all"
    """
    if root_env + "_SINGLE" in os.environ:
        path = os.environ[root_env + "_SINGLE"]
    elif root_env in os.environ:
        path = os.path.join(os.environ[root_env], observatory)
    elif "CRDS_PATH_SINGLE" in os.environ:
        path = os.path.join(os.environ["CRDS_PATH_SINGLE"], subdir)
    elif "CRDS_PATH" in os.environ:
        path = os.path.join(os.environ["CRDS_PATH"], subdir, observatory)
    else:
        path = os.path.join(CRDS_DEFAULT_CACHE, subdir, observatory)
    return _clean_path(path)

def get_crds_path():
    """
    >>> temp = dict(os.environ)
    >>> os.environ = {}

    >>> get_crds_path() == CRDS_DEFAULT_CACHE
    True

    >>> os.environ = {}
    >>> os.environ["CRDS_PATH_SINGLE"] = "/somewhere"
    >>> get_crds_path()
    '/somewhere'

    >>> os.environ = {}
    >>> os.environ["CRDS_PATH"] = "/somewhere_else"
    >>> get_crds_path()
    '/somewhere_else'

    >>> os.environ = temp
    """
    return _std_cache_path("", "*not-in-crds-environment*", "")

def get_crds_mappath(observatory):
    """get_crds_mappath() returns the base path of the CRDS mapping directory
    tree where CRDS rules files (mappings) are stored.   This is extended by
    <observatory> once it is known.

    DEPRECATED:  only use in the config module.  Use locate_file() or locate_mapping() instead.

    >>> temp = dict(os.environ)
    >>> os.environ = {}

    >>> get_crds_mappath('jwst') == CRDS_DEFAULT_CACHE + "/mappings/jwst"
    True

    >>> os.environ["CRDS_PATH"] = '/somewhere'
    >>> get_crds_mappath('jwst')
    '/somewhere/mappings/jwst'

    >>> os.environ["CRDS_PATH_SINGLE"] = "/somewhere2"
    >>> get_crds_mappath('jwst')
    '/somewhere2/mappings'

    >>> os.environ["CRDS_MAPPATH"] = "/somewhere3/mappings2"
    >>> get_crds_mappath('jwst')
    '/somewhere3/mappings2/jwst'

    >>> os.environ["CRDS_MAPPATH_SINGLE"] = "/somewhere4/mappings3"
    >>> get_crds_mappath('jwst')
    '/somewhere4/mappings3'

    >>> os.environ = temp
    """
    return _std_cache_path(observatory, "CRDS_MAPPATH", "mappings")

def get_crds_cfgpath(observatory):
    """Return the path to a writable directory used to store configuration info
    such as last known server status.   This is extended by <observatory> once
    it is known.   If CRDS_PATH doesn't point to a writable directory, then
    CRDS_CFGPATH should be defined.
    """
    return _std_cache_path(observatory, "CRDS_CFGPATH", "config")

def get_crds_root_cfgpath():
    """Return the root multi-project config path. Used for e.g. lockfiles.

    >>> temp = dict(os.environ)
    >>> os.environ = {}

    >>> os.environ["CRDS_PATH"] = '/somewhere'
    >>> get_crds_root_cfgpath()
    '/somewhere/config'

    >>> os.environ["CRDS_PATH_SINGLE"] = "/somewhere2"
    >>> get_crds_root_cfgpath()
    '/somewhere2/config'

    >>> os.environ["CRDS_CFGPATH"] = "/somewhere3/cfg2"
    >>> get_crds_root_cfgpath()
    '/somewhere3/cfg2'

    >>> os.environ["CRDS_CFGPATH_SINGLE"] = "/somewhere4/cfg3"
    >>> get_crds_root_cfgpath()
    '/somewhere4/cfg3'

    >>> os.environ = temp
    """
    dirname = get_crds_cfgpath("all")
    if dirname.endswith("all"):
        return os.path.dirname(dirname)
    else:
        return dirname

def get_crds_refpath(observatory):
    """get_crds_refpath returns the base path of the directory tree where CRDS
    reference files are stored.   This is extended by <observatory> once it is
    known.

    DEPRECATED: only use in config module.  Use locate_file() or locate_reference() instead.
    """
    return _std_cache_path(observatory, "CRDS_REFPATH", "references")

def locate_config(cfg, observatory):
    """Return the absolute path where reference `cfg` should be located."""
    if os.path.dirname(cfg):
        return cfg
    return os.path.join(get_crds_cfgpath(observatory), cfg)

def override_crds_paths(crds_path):
    """Overrides all CRDS PATH definition environment variables
    (e.g. CRDS_MAPPATH, CRDS_MAPPATH_SINGLE, CRDS_REFPATH, ...)
    as if the user had only specified CRDS_PATH=crds_path and let
    all other cache subpaths be defined at default locations.
    """
    for name in os.environ:
        if "CRDS_" in name and "PATH" in name:
            del os.environ[name]
    os.environ["CRDS_PATH"] = crds_path


# -------------------------------------------------------------------------------------

def get_crds_picklepath(observatory):
    """Return the directory name where CRDS stores pickles for `observatory`."""
    return _std_cache_path(observatory, "CRDS_PICKLEPATH", "pickles")

def locate_pickle(mapping, observatory=None):
    """Return the absolute path where reference `ref` should be located."""
    if os.path.dirname(mapping):
        return mapping
    if observatory is None:
        observatory = mapping_to_observatory(mapping)
    return os.path.join(get_crds_picklepath(observatory), mapping + ".pkl")

USE_PICKLED_CONTEXTS = BooleanConfigItem("CRDS_USE_PICKLED_CONTEXTS", False,
    "When True,  CRDS contexts should be loaded from a pickled version if possible.")

AUTO_PICKLE_CONTEXTS = BooleanConfigItem("CRDS_AUTO_PICKLE_CONTEXTS", False,
    "When True, CRDS contexts should be automatically pickled and cached after loading.")

# -------------------------------------------------------------------------------------

FORCE_COMPLETE_LOAD = BooleanConfigItem("CRDS_FORCE_COMPLETE_LOAD", False,
    "When True, force CRDS contexts to load in their entirety rather than based on what is actually used.")

EXPLICIT_GARBAGE_COLLECTION = BooleanConfigItem("CRDS_EXPLICIT_GARBAGE_COLLECTION", True,
    "When False, the @gc_collected function decorator skips garbage collection.")
# -------------------------------------------------------------------------------------

def get_sqlite3_db_path(observatory):
    """Return the path to the downloadable CRDS catalog + history SQLite3 database file."""
    return locate_config("crds_db.sqlite3", observatory)

# ===========================================================================

CRDS_SUBDIR_TAG_FILE = "ref_cache_subdir_mode"
CRDS_REF_SUBDIR_MODES = ["instrument", "flat", "legacy"]
CRDS_REF_SUBDIR_MODE = "None"  # does not make cache consistent with env,  test only!!

def get_crds_ref_subdir_mode(observatory):
    """Return the mode value defining how reference files are located."""
    global CRDS_REF_SUBDIR_MODE
    if CRDS_REF_SUBDIR_MODE in ["None", None]:
        mode_path = get_crds_ref_subdir_file(observatory)
        try:
            with open(mode_path) as pfile:
                mode = pfile.read().strip()
            # log.verbose("Determined cache format from", repr(mode_path), "as", repr(mode))
        except IOError:
            if len(glob.glob(os.path.join(get_crds_refpath(observatory), "*"))) > 20:
                mode = "flat"
                log.verbose("No cache config tag found, looks like a 'flat' cache based on existing references.")
            else:
                mode = "instrument"
                log.verbose("No cache config tag found, defaulting to 'instrument' based cache.")
            with log.verbose_on_exception("Failed saving default subdir mode to", repr(mode)):
                set_crds_ref_subdir_mode(mode, observatory)
        check_crds_ref_subdir_mode(mode)
        CRDS_REF_SUBDIR_MODE = mode
    else:
        mode = CRDS_REF_SUBDIR_MODE
    return mode

def set_crds_ref_subdir_mode(mode, observatory):
    """Set the reference file location subdirectory `mode`."""
    global CRDS_REF_SUBDIR_MODE
    check_crds_ref_subdir_mode(mode)
    CRDS_REF_SUBDIR_MODE = mode
    mode_path = get_crds_ref_subdir_file(observatory)
    if writable_cache_or_verbose("skipping subdir mode write."):
        from crds.core import heavy_client as hv   #  yeah,  kinda gross
        hv.cache_atomic_write(mode_path, mode, "Couldn't save sub-directory mode config")

def get_crds_ref_subdir_file(observatory):
    """Return path to CRDS config file defining cache organization.  Created
    automatically,  this is nominally for CRDS internal use only and cleanup.

    Returns <path to cache organization config file>
    """
    return os.path.join(get_crds_cfgpath(observatory), CRDS_SUBDIR_TAG_FILE)


def check_crds_ref_subdir_mode(mode):
    """Check for valid reference location subdirectory `mode`."""
    assert mode in CRDS_REF_SUBDIR_MODES, "Invalid CRDS cache subdirectory mode = " + repr(mode)
    return mode

# ===========================================================================

CRDS_MODE = StrConfigItem("CRDS_MODE", default="auto",
    comment="""Selects where bestrefs are performed, locally, remotely (on the CRDS server),
    or automatically chosen by comparing client version to server version.""",
    valid_values=["local", "remote", "auto"])

def get_crds_processing_mode():
    """Return the preferred location for computing best references when
    network connectivity is available.

    'local'   --   compute locally even if client CRDS is obsolete
    'remote'  --   compute remotely even if client CRDS is up-to-date
    'auto'    --   compute locally unless connected and client CRDS is obsolete
    """
    return CRDS_MODE.get()

# ===============================================================================================

# CRDS_CONTEXT = StrConfigItem("CRDS_CONTEXT", default=None)

def get_crds_env_context():
    """If it has been specified in the environment by CRDS_CONTEXT,  return the
    pipeline context which defines CRDS best reference rules,  else None.

    >>> os.environ["CRDS_CONTEXT"] = "jwst.pmap"
    >>> get_crds_env_context()
    'jwst.pmap'

    >>> os.environ["CRDS_CONTEXT"] = "jwst_miri_0022.imap"
    >>> get_crds_env_context()
    Traceback (most recent call last):
    ...
    AssertionError: Only set CRDS_CONTEXT to a literal or symbolic context (.pmap), e.g. jwst_0042.pmap,  jwst-2014-10-15T00:15:21, jwst-operational,  not 'jwst_miri_0022.imap'

    >>> del os.environ["CRDS_CONTEXT"]
    >>> get_crds_env_context()
    """
    context = os.environ.get("CRDS_CONTEXT", None)
    if context:
        assert is_context_spec(context), \
            "Only set CRDS_CONTEXT to a literal or symbolic context (.pmap), e.g. jwst_0042.pmap,  jwst-2014-10-15T00:15:21, jwst-operational,  not " + repr(context)
    return context

CRDS_IGNORE_MAPPING_CHECKSUM = BooleanConfigItem("CRDS_IGNORE_MAPPING_CHECKSUM", False,
    "disables mapping checksums during development.")

def get_ignore_checksum():
    """Returns environment override for disabling mapping checksums during development."""
    return env_to_bool("CRDS_IGNORE_MAPPING_CHECKSUM", False)

def get_log_time():
    """Returns override flag for outputting time in log messages."""
    return env_to_bool("CRDS_LOG_TIME", False)

# ===========================================================================

def get_crds_env_vars():
    """Return a dictionary of all env vars starting with 'CRDS'."""
    env_vars = {}
    for var in os.environ:
        if var.upper().startswith("CRDS"):
            env_vars[var] = os.environ[var]
    return env_vars

def get_crds_actual_paths(observatory):
    """Return a dictionary of the critical paths CRDS configuration resolves to."""
    return {
        "mapping root" : get_crds_mappath(observatory),
        "reference root" : get_crds_refpath(observatory),
        "config root" : get_crds_cfgpath(observatory),
        "pickle root" : get_crds_picklepath(observatory),
        }

# ============================================================================

# client API related settings.

CRDS_DOWNLOAD_MODE = StrConfigItem("CRDS_DOWNLOAD_MODE", "http",
    "Selects the mode used by the CRDS client for downloading rules and references.",
    ["http", "plugin"], lower=True)

def get_download_mode():
    """Return the mode used to download references and mappings,  either normal
    "http" file transfer or "plugin" based.   "rpc" mode has been removed.

    Returns "http" or "plugin"
    """
    return CRDS_DOWNLOAD_MODE.get()

def get_download_plugin():
    """Fetch a command template from the environment to use a as a substitute for CRDS
    built-in downloaders.   This can be used to apply "wget" or "curl", etc, to perform
    downloads as sub-processes rather than as a direct Python http implementation.
    """
    if "CRDS_DOWNLOAD_MODE" in os.environ and os.environ["CRDS_DOWNLOAD_MODE"].lower() != "plugin":
        return None
    elif "CRDS_DOWNLOAD_PLUGIN" in os.environ:
        return os.environ["CRDS_DOWNLOAD_PLUGIN"]
    elif get_download_mode() == "plugin":
        for path in ["/usr/local/bin", "/usr/bin", "/sw/bin"]:
            program = os.path.join(path, "wget")
            if os.path.exists(program):
                return program + " --no-check-certificate --quiet ${SOURCE_URL}  -O ${OUTPUT_PATH}"
    return None

# -------------------------------------------------------------------------------------

# This permits an AWS environment to override the URLs normally supplied by the
# CRDS server with locally defined (S3) source URLs.  This can keep actual file
# downloads local to AWS and scalable with S3.

CRDS_CONFIG_URI = StrConfigItem("CRDS_CONFIG_URI", None,
    "Defines the URL/URI from which CRDS config files are downloaded.",
    lower=True)

CRDS_MAPPING_URI = StrConfigItem("CRDS_MAPPING_URI", None,
    "Defines the URL/URI from which CRDS mapping files are downloaded.",
    lower=True)

CRDS_REFERENCE_URI = StrConfigItem("CRDS_REFERENCE_URI", None,
    "Defines the URL/URI from which CRDS reference files are downloaded.",
    lower=True)

CRDS_PICKLE_URI = StrConfigItem("CRDS_PICKLE_URI", None,
    "Defines the URL/URI from which CRDS pickle files are downloaded.",
    lower=True)

# -------------------------------------------------------------------------------------

DOWNLOAD_CHECKSUMS = BooleanConfigItem(
    "CRDS_DOWNLOAD_CHECKSUMS", True, "Verify downloaded files match server's sha1sum.  If false,  allow bad checksums.")

DOWNLOAD_LENGTHS = BooleanConfigItem(
    "CRDS_DOWNLOAD_LENGTHS", True, "Verify downloaded files match server's file length.  If false,  allow bad lengths.")

def get_checksum_flag():
    """Return True if the environment is configured for checksums."""
    return DOWNLOAD_CHECKSUMS.get()

def get_length_flag():
    """Return True if the environment is configured for verifying downloaded file lengths.
    """
    return DOWNLOAD_LENGTHS.get()

# -------------------------------------------------------------------------------------

CLIENT_RETRY_COUNT = IntConfigItem(
    "CRDS_CLIENT_RETRY_COUNT", 1, "Integer number of times CRDS should retry download errors.  No retries == 1.")

def get_client_retry_count():
    """Return the integer number of times a network transaction should be attempted.  No retries == 1."""
    return CLIENT_RETRY_COUNT.get()

CLIENT_RETRY_DELAY_SECONDS = IntConfigItem(
    "CRDS_CLIENT_RETRY_DELAY_SECONDS", 0, "Seconds CRDS should wait between retries.  Defaults to 0.")

def get_client_retry_delay_seconds():
    """Return the integer number of seconds CRDS should wait between retrying failed network transactions."""
    return CLIENT_RETRY_DELAY_SECONDS.get()

def enable_retries(retry_count=20, delay_seconds=10):
    """Set reasonable defaults for CRDS retries"""
    CLIENT_RETRY_COUNT.set(retry_count)
    CLIENT_RETRY_DELAY_SECONDS.set(delay_seconds)

def disable_retries():
    """Set the defaults for only one try for each network transaction."""
    CLIENT_RETRY_COUNT.set(1)
    CLIENT_RETRY_DELAY_SECONDS.set(0)

# -------------------------------------------------------------------------------------

CRDS_DEFAULT_SERVERS = {
    "hst" : "https://hst-crds.stsci.edu",
    "jwst" : "https://jwst-crds.stsci.edu",
    "roman" : "https://roman-crds.stsci.edu",
    # None : "https://crds-serverless-mode.stsci.edu"
}

def get_server_url(observatory):
    """Return either the value of CRDS_SERVER_URL or an appropriate default server for `observatory`."""
    return os.environ.get("CRDS_SERVER_URL", CRDS_DEFAULT_SERVERS.get(observatory, None))

# ===========================================================================

USERNAME = None

def get_username(override=None):
    """Initialize the USERNAME config item and return the value, setting it to
    `override` is override is not None.  (nominally command line override).
    Defer username initialization until requested to avoid unneeded dialogs.
    """
    global USERNAME
    if USERNAME is None:
        # get env_var or ini_file
        USERNAME = StrConfigItem(
            "CRDS_USERNAME", None, ini_section="authentication",
            comment="User's e-mail on CRDS server, defaulting to current login.")
        # ultimately: override or env_var or ini_file or fallback function
        username = override or USERNAME.get()
        if username in ["None", "none", None]:
            username = getpass.getuser()
        USERNAME.set(username)
    return USERNAME.get()

PASSWORD = None

def get_password(override=None):
    """Initialize the PASSWORD config item and return the value setting it to
    `override` if override is not None.  (nominally command line override.)
    Defer password initialization until requested to avoid unneeded dialogs.
    """
    global PASSWORD
    if PASSWORD is None:
        # ultimately: override or env_var or ini_file or fallback function
        PASSWORD = StrConfigItem(
            "MAST_API_TOKEN", None, ini_section="authentication",
            comment="User's MAST_API_TOKEN, defaulting to echo-less key entry.")
        password = override or PASSWORD.get()
        if password in ["None", "none", None]:
            password = getpass.getpass("MAST_API_TOKEN: ")
        PASSWORD.set(password)
    return PASSWORD.get()

# ===========================================================================

_CRDS_CACHE_READONLY = False

def set_cache_readonly(readonly=True):
    """Set the flag controlling writes to the CRDS cache."""
    global _CRDS_CACHE_READONLY
    _CRDS_CACHE_READONLY = readonly
    return _CRDS_CACHE_READONLY

def get_cache_readonly():
    """Read the flag controlling writes to the CRDS cache.  When True,  no write to cache should occur."""
    return _CRDS_CACHE_READONLY or env_to_bool("CRDS_READONLY_CACHE", False)

def writable_cache_abort(func):
    """Generator a filter by decorating `func`.  These call `func` and return False when the CRDS cache is
    readonly,  otherwise they return True.

    Use like this:

    if writable_cache_or_info(... log message parameters...):
        block of code requiring writable cache
    """
    def func_check_writable(*args, **keys):
        """func_check_writable is a wrapper which issues a func() message when CRDS
        is configured for a readonly cache.
        """
        if get_cache_readonly():  # message and quit
            func("READONLY CACHE", *args, **keys)
            return False
        else:
            return True
        func_check_writable.__name__ = "wrapped_writable_" + func.__name__
        func_check_writable._wrapped_writable = True
    return func_check_writable

writable_cache_or_info = writable_cache_abort(log.info)
writable_cache_or_verbose = writable_cache_abort(log.verbose)
writable_cache_or_warning = writable_cache_abort(log.warning)

# ===========================================================================

def filetype(filename):
    """Classify `filename`'s type so it can be processed or displayed.

    >>> filetype("foo.fits")
    'fits'

    >>> filetype("foo.yaml")
    'yaml'

    >>> filetype("foo.pmap")
    'mapping'

    >>> filetype("foo.json")
    'json'

    >>> filetype("foo.txt")
    'text'

    >>> filetype("foo.r0h")
    'geis'

    >>> filetype("foo.r0d")
    'geis'

    >>> filetype('foo.exe')
    'unknown'
    """
    if is_mapping(filename):
        return "mapping"
    elif filename.endswith(".fits"):
        return "fits"
    elif filename.endswith(".yaml"):
        return "yaml"
    elif filename.endswith(".json"):
        return "json"
    elif filename.endswith(".asdf"):
        return "asdf"
    elif filename.endswith(".txt"):
        return "text"
    elif re.match(r".*\.r[0-9][hd]$", filename): # GEIS header
        return "geis"
    else:
        return "unknown"

def file_in_cache(filename, observatory):
    """Return True IFF `filename` is in the local cache."""
    path = locate_file(os.path.basename(filename), observatory)
    return os.path.exists(path)

# ===========================================================================

def get_path(filename, observatory):
    """Return the CRDS cache path of `filename` for `observatory`.   The path is
    similar to the one from locate_file(),  but includes dirname only,  not `filename` itself.
    """
    assert OBSERVATORY_RE.match(observatory), "Invalid observatory " + repr(observatory)
    fullpath = locate_file(filename, observatory)
    return os.path.dirname(fullpath)

def locate_file(filepath, observatory):
    """Returns CRDS cache path if `filepath` has no directory, otherwise `filepath` as-is.

   Cannot always determine CRDS cache location for hypothetical files if not
   sufficiently identified in basename of `filepath`.

   Used to determine cache locations for directory-less `filepaths` from CRDS rules.

   Returns complex filepaths as-is for supporting operations on external files.
    """
    if os.path.dirname(filepath):
        return filepath
    return relocate_file(filepath, observatory)

def pop_crds_uri(filepath):
    """Pop off crds:// from a filepath,  yielding a pathless filename."""
    if filepath.startswith("crds://"):
        filepath = filepath[len("crds://"):]
        assert not os.path.dirname(filepath), "crds:// must prefix a filename with no path."
    return filepath

def relocate_file(filepath, observatory):
    """Returns path in CRDS cache where `filepath` would be relocated if it were
    copied into the CRDS cache.

    Able to determine CRDS cache location based on filename or file
    contents when `filepath` points to a real file.
    """
    # IMPORTANT:  since the outcome of relocate_file is ALWAYS a CRDS cache path,
    # the "dirname alresady defined" short-cut should not be used here.  The existing
    # dirname is irrelevant execept for determining file properties from badly named
    # reference files by inspecting the header.
    if is_mapping(filepath):
        return relocate_mapping(filepath, observatory)
    else:
        return relocate_reference(filepath, observatory)

# ===========================================================================

def locate_reference(ref, observatory):
    """Returns CRDS cache path for `ref` if it specifies no directory, otherwise
    `ref` as-is.
    """
    if os.path.dirname(ref):
        return ref
    return relocate_reference(ref, observatory)

def relocate_reference(ref, observatory):
    """Returns CRDS cache location where `ref` would be copied if it
    was copied into the CRDS cache.  When `ref` specifies a path to an
    existing file, the contents of `ref` can be exploited to determine
    the CRDS cache location.  Otherwise, the basename of `ref` must be
    in a standard form which implies location.
    """
    # This limited case is required for the server and dealing with temporary filenames
    # which cannot be used to determine instrument subdirectory based on name.
    if ref == "N/A":
        return ref
    if get_crds_ref_subdir_mode(observatory) == "flat":
        return os.path.join(get_crds_refpath(observatory), os.path.basename(ref))
    else:
        from crds.core import utils
        return utils.get_locator_module(observatory).locate_file(ref)

# ===========================================================================
if os.path.exists("/tmp"):
    DEFAULT_LOCK_PATH = "/tmp"
else:
    DEFAULT_LOCK_PATH = os.path.join(get_crds_root_cfgpath(), "locks")

CACHE_LOCK_PATH = StrConfigItem(
    "CRDS_LOCK_PATH", DEFAULT_LOCK_PATH,
    "Lock directory used to store lock files,  nominally for CRDS cache.")

USE_LOCKING = BooleanConfigItem("CRDS_USE_LOCKING", True,
    "Set to False to turn off CRDS cache locking.")

LOCKING_MODE = StrConfigItem("CRDS_LOCKING_MODE",  "multiprocessing",
    "Form of locking used by CRDS cache.",
    valid_values=["lockfile", "filelock", "multiprocessing"],
    lower=True)

def get_crds_lockpath(lock_filename):
    """Return the full path of `lock_filename` filename based on CRDS lock path configuration."""
    return os.path.join(CACHE_LOCK_PATH.get(), lock_filename)

# ===========================================================================

def complete_re(regex_str):
    """Add ^$ to `regex_str` to force match to entire string."""
    return "^" + regex_str + "$"

OBSERVATORY_RE_STR = r"[a-zA-Z]{1,8}"
OBSERVATORY_RE = re.compile(complete_re(OBSERVATORY_RE_STR))
FILE_RE_STR = r"[A-Za-z0-9\._\- ]{1,128}|N/A|OMIT"
FILE_RE = re.compile(complete_re(FILE_RE_STR)) # at min *should not* contain % < > \ { }
FILE_PATH_RE_STR = r"([/A-Za-z0-9_\- ]{1,256})?" + FILE_RE_STR
FILE_PATH_RE = re.compile(complete_re(FILE_PATH_RE_STR))  # at min *should not* contain % < > \ { }

def check_filename(filename):
    """Verify that `filename` is a basename with no dangerous characters."""
    assert FILE_RE.match(filename), "Invalid file name " + repr(filename)

def check_path(path):
    """Make sure `path` has no dangerous characters."""
    path = os.path.abspath(path)
    assert FILE_PATH_RE.match(path), "Invalid file path " + repr(path)
    return path

# -------------------------------------------------------------------------------------

def is_reference(reference):
    """Return True IFF file name `reference` is plausible as a reference file name.
    is_reference() does not *guarantee* that `reference` is a reference file name,
    in particular a dataset filename might pass as a reference.

    >>> is_reference("something.fits")
    True
    >>> is_reference("something.asdf")
    True
    >>> is_reference("something.json")
    True
    >>> is_reference("something.yaml")
    True
    >>> is_reference("something.text")
    True
    >>> is_reference("something.r0h")
    True
    >>> is_reference("something.r0d")
    True
    >>> is_reference("something.foo")
    False
    >>> is_reference("/some/path/something.fits")
    True
    >>> is_reference("/some/path/something.pmap")
    False
    >>> is_reference("/some/path/something.imap")
    False
    >>> is_reference("something.rmap")
    False
    """
    extension = os.path.splitext(reference)[-1].lower()
    return bool(re.match(r"\.fits|\.asdf|\.r\d[hd]|\.yaml|\.json|\.text", extension))

# -------------------------------------------------------------------------------------

# max len name component == 32.  max components == 3.
# no digits in CRDS name components,  except serial no
CRDS_OBS_RE_STR = r"(?P<observatory>" + r"[a-z][a-z]{1,8})" # No hyphens or digits
CRDS_INSTR_RE_STR = r"(?P<instrument>" + r"[a-z][a-z0-9]{1,16})"   # No hyphens
CRDS_FILEKIND_RE_STR = r"(?P<filekind>" + r"[a-z][a-z0-9\-]{1,32})"
CRDS_BASE_NAME_RE_STR = (r"(" +
        CRDS_OBS_RE_STR + r"(_" +
            CRDS_INSTR_RE_STR + r"(_" +
                CRDS_FILEKIND_RE_STR +
            r")?" +
        r")?" +
    r")")

CRDS_SYM_NAME_RE_STR = (r"(" +
        CRDS_OBS_RE_STR + r"(\-" +
            CRDS_INSTR_RE_STR + r"(\-" +
                CRDS_FILEKIND_RE_STR +
            r")?" +
        r")?" +
    r")")

CRDS_NAME_RE_STR = CRDS_BASE_NAME_RE_STR + r"(_\d\d\d\d)?\."
CRDS_NAME_RE = re.compile(CRDS_NAME_RE_STR)   # intentionally not complete. no ^ or $

# s7g1700gl_dead.fits
CDBS_NAME_RE_STR = r"[a-z0-9_]{1,52}\.(fits|r\d[hd])"
CDBS_NAME_RE = re.compile(complete_re(CDBS_NAME_RE_STR))

# -------------------------------------------------------------------------------------

def is_valid_reference_name(filename):
    """Return True IFF `filename` has a valid CRDS reference filename format.

    >>> is_valid_reference_name("hst_acs_darkfile_0027.rmap")
    False

    >>> is_valid_reference_name("hst_acs_darkfile_0027.fits")
    True

    >>> is_valid_reference_name("s7g1700gl_dead.fits")
    True

    >>> is_valid_reference_name("jwst_miri_pars-tweakreg_0001.asdf")
    True
    """
    name = os.path.basename(filename)
    return is_reference(name) and (is_crds_name(name) or is_cdbs_name(name))

def is_crds_name(name):
    """Return True IFF `name` is a valid CRDS-style name.

    >>> is_crds_name("hst_acs_darkfile_0027.rmap")
    True

    >>> is_crds_name("hst_acs.imap")
    True

    >>> is_crds_name("hst_acs_darkfile_0027.fits")
    True

    >>> is_crds_name("s7g1700gl_dead.fits")
    False

    >>> is_crds_name("jwst_miri_pars-tweakreg_0001.asdf")
    True

    >>> is_crds_name("jwst_miri_pars-tweakreg_0001.rmap")
    True
    """
    name = os.path.basename(name).lower()
    return bool(CRDS_NAME_RE.match(name))

def is_cdbs_name(name):
    """Return True IFF `name is a valid CDBS-style name.

    >>> is_cdbs_name("hst_acs_darkfile_0027.rmap")
    False

    >>> is_cdbs_name("s7g1700gl_dead.fits")
    True

    >>> is_cdbs_name("jwst_miri_pars-tweakreg_0001.asdf")
    False
    """
    name = os.path.basename(name).lower()
    return bool(CDBS_NAME_RE.match(name))

# -------------------------------------------------------------------------------------

# Standard date time format using T separator for command line use specifying contexts.
# e.g. 2040-02-22T12:01:30.4567
CONTEXT_DATETIME_RE_STR = (
    r"\d\d\d\d\-\d\d\-\d\d" +
        r"(T\d\d:\d\d:\d\d" +
            r"(\.\d+)?" +
        r")?"
    )

CONTEXT_DATETIME_RE = re.compile(complete_re(CONTEXT_DATETIME_RE_STR))

CONTEXT_OBS_RE_STR = r"[a-z]{1,8}"

# e.g.   2040-02-22T12:01:30.4567,  hst-2040-02-22T12:01:30.4567, hst-acs-2040-02-22T12:01:30.4567, ...
CONTEXT_RE_STR = (
    r"(" +
        r"(?P<context>" + CRDS_SYM_NAME_RE_STR + ")" +
        r"(" +
            r"(\-" +
                r"(?P<date>" +
                    r"(" + CONTEXT_DATETIME_RE_STR + r")" +
                        r"|" +
                    r"(" + "edit|operational|versions" + r")" +
                r")" +
            r")" +
        r")" +
    r")" +
        r"|" +
    r"(" +
        r"(" + CONTEXT_DATETIME_RE_STR + r")" +
    r")"
    )
CONTEXT_RE = re.compile(complete_re(CONTEXT_RE_STR))

PIPELINE_CONTEXT_RE_STR = (
    r"(" +
        r"(?P<context>" + CRDS_OBS_RE_STR + ")" +
        r"(" +
            r"(\-" +
                r"(" +
                    r"(?P<date>" + CONTEXT_DATETIME_RE_STR + r")" +
                        "|" +
                    r"(?P<context_tag>" + "edit|operational|versions" + r")" +
                r")" +
            r")" +
        r")"
    r")" +
        r"|" +
    r"(" +
        r"(" + CONTEXT_DATETIME_RE_STR + r")" +
    r")"
    )
PIPELINE_CONTEXT_RE = re.compile(complete_re(PIPELINE_CONTEXT_RE_STR))

MAPPING_RE_STR = CRDS_NAME_RE_STR + r"[pir]map"
MAPPING_RE = re.compile(complete_re(MAPPING_RE_STR))

# -------------------------------------------------------------------------------------

def is_mapping(mapping):
    """Return True IFF `mapping` has an extension indicating a CRDS mapping file."""
    return isinstance(mapping, str) and mapping.endswith((".pmap", ".imap", ".rmap"))

def is_simple_crds_mapping(mapping):
    """Return True IFF `mapping` implicitly or explicitly exists in the CRDS cache.

    Return False for ad hoc mappings outside the CRDS Cache.

    Nominally this function is one determinant of whether or not a mapping is
    picklable at the level of the overall CRDS runtime system.

    >>> is_simple_crds_mapping(locate_mapping("hst.pmap"))
    True

    is_simple_crds_mapping() is not a full guarantee that a mapping actually exists in
    CRDS,  merely a guarantee that if it did it would be located implicitly in
    the CRDS cache and has a basic name format which is not a date based context
    specifier.

    >>> is_simple_crds_mapping("hst_foo.pmap")
    True

    >>> is_simple_crds_mapping(locate_mapping("hst.fits"))
    False

    >>> is_simple_crds_mapping("tests/data/hst.pmap")
    False

    >>> is_simple_crds_mapping("hst-2040-01-29T12:00:00")
    False

    >>> is_simple_crds_mapping("hst-acs-2040-01-29T12:00:00")
    False

    >>> is_simple_crds_mapping("hst-acs-darkfile-2040-01-29T12:00:00")
    False

    >>> is_simple_crds_mapping("2040-01-29T12:00:00")
    False

    >>> is_simple_crds_mapping("hst-edit")
    False

    >>> is_simple_crds_mapping("hst-cos-deadtab-edit")
    False

    >>> is_simple_crds_mapping("jwst-operational")
    False

    >>> is_simple_crds_mapping("hst-foo")
    False
    """
    if isinstance(mapping, str):
        basename = os.path.basename(mapping)
        return is_valid_mapping_name(basename) and (locate_mapping(mapping) == locate_mapping(basename))
    else:
        return False

def is_valid_mapping_name(mapping):
    """Return True IFF `mapping` has a CRDS-style root name and a mapping extension."""
    return is_mapping(mapping) and bool(MAPPING_RE.match(mapping))

def is_mapping_spec(mapping):
    """Return True IFF `mapping` is a mapping name *or* a date based mapping specification.

    Date-based specifications can be interpreted by the CRDS server with respect to the operational
    context history to determine the default operational context which was in use at that date.
    This function verifies syntax only,  not the existence of corresponding context.

    >>> is_mapping_spec("hst.pmap")
    True

    >>> is_mapping_spec("foo.pmap")
    True

    >>> is_mapping_spec("foo")
    False

    >>> is_mapping_spec("hst-2040-01-29T12:00:00")
    True

    >>> is_mapping_spec("hst-acs-2040-01-29T12:00:00")
    True

    >>> is_mapping_spec("jwst-miri-pars-tweakreg-2040-01-29T12:00:00")
    True

    >>> is_mapping_spec("jwst-miri-pars-tweakreg-edit")
    True

    >>> is_mapping_spec("2040-01-29T12:00:00")
    True

    >>> is_mapping_spec("hst-edit")
    True

    >>> is_mapping_spec("hst-cos-deadtab-edit")
    True

    >>> is_mapping_spec("jwst-operational")
    True

    >>> is_mapping_spec("hst-foo")
    False

    >>> is_mapping_spec("hst_wfc3_0001.imap")
    True
    """
    return is_mapping(mapping) or (isinstance(mapping, str) and bool(CONTEXT_RE.match(mapping)))

def is_context(mapping):
    """Return True IFF `mapping` has an extension indicating a CRDS CONTEXT, i.e. .pmap."""
    return isinstance(mapping, str) and mapping.endswith((".pmap",))

def is_context_spec(mapping):
    """Return True IFF `mapping` is a mapping name *or* a date based mapping specification.

    Date-based specifications can be interpreted by the CRDS server with respect to the operational
    context history to determine the default operational context which was in use at that date.
    This function verifies syntax only,  not the existence of corresponding context.

    >>> is_context_spec("hst_0042.pmap")
    True

    >>> is_context_spec("hst.pmap")
    True

    >>> is_context_spec("foo.pmap")
    True

    >>> is_context_spec("foo")
    False

    >>> is_context_spec("hst-edit")
    True

    >>> is_context_spec("hst-2040-01-29T12:00:00")
    True

    >>> is_context_spec("hst-acs-2040-01-29T12:00:00")
    False
    """
    return is_context(mapping) or (isinstance(mapping, str) and bool(PIPELINE_CONTEXT_RE.match(mapping)))

def is_date_based_mapping_spec(mapping):
    """Return True IFF `mapping` is a date based specification (not a filename).

    >>> is_date_based_mapping_spec("2040-01-29T12:00:00")
    True

    >>> is_date_based_mapping_spec("jwst-miri-pars-tweakreg-2040-01-29T12:00:00")
    True

    >>> is_date_based_mapping_spec("hst.pmap")
    False
    """
    return is_mapping_spec(mapping) and not is_mapping(mapping)

def locate_mapping(mappath, observatory=None):
    """Return the CRDS cache path for mapping `mappath` if it has no directory,
    otherwise returns mappath as-is.
    """
    if os.path.dirname(mappath):
        return mappath
    return relocate_mapping(mappath, observatory)

def relocate_mapping(mappath, observatory=None):
    """Return the CRDS cache path where CRDS mapping `mappath` should be
    be located,  replacing the path in `mappath` with the cache path.
    """
    if observatory is None:
        observatory = mapping_to_observatory(mappath)
    if mappath == "N/A":
        return mappath
    return os.path.join(get_crds_mappath(observatory), os.path.basename(mappath))

def mapping_exists(mapping):
    """Return True IFF `mapping` exists on the local file system."""
    return os.path.exists(locate_mapping(mapping))

# These are name based but could be written as slower check-the-mapping
# style functions.
def mapping_to_observatory(context_file):
    """
    >>> mapping_to_observatory('hst.pmap')
    'hst'
    >>> mapping_to_observatory('hst_0000.pmap')
    'hst'
    >>> mapping_to_observatory('hst_acs_biasfile_0000.rmap')
    'hst'
    """
    return os.path.basename(context_file).split("_")[0].split(".")[0]

def mapping_to_instrument(context_file):
    """
    >>> mapping_to_instrument('hst_acs_biasfile.rmap')
    'acs'
    """
    return os.path.basename(context_file).split("_")[1].split(".")[0]

def mapping_to_filekind(context_file):
    """
    >>> mapping_to_filekind('hst_acs_biasfile.rmap')
    'biasfile'
    """
    return os.path.basename(context_file).split("_")[2].split(".")[0]

# -------------------------------------------------------------------------------------
def is_config(filename):
    """Return True IFF `filename` names a CRDS configuration file."""
    filename = os.path.basename(filename)
    if filename in ["server_config"]:
        return True
    return False

# -------------------------------------------------------------------------------------
def is_pickle(filename):
    """Return True IFF `filename` names a CRDS pickle file."""
    filename = os.path.basename(filename)
    if filename.endswith(".pkl"):
        return True
    return False

# -------------------------------------------------------------------------------------

def get_crds_state():
    """Capture the current CRDS configuration and return it as a dictionary.
    Intended for customizing state during self-tests and restoring during teardown.
    """
    from .log import get_verbose
    env = { key : val for key, val in os.environ.items() if key.startswith("CRDS_") }
    env["CRDS_REF_SUBDIR_MODE"] = CRDS_REF_SUBDIR_MODE
    env["_CRDS_CACHE_READONLY"] = get_cache_readonly()
    env["PASS_INVALID_VALUES"] = PASS_INVALID_VALUES.get()
    env["CRDS_VERBOSITY"] = get_verbose()
    return env

def set_crds_state(old_state):
    """Restore the configuration of CRDS returned by get_crds_state()."""
    from crds.client import api   # deferred circular import
    from .log import set_verbose
    # determination of observatory and server URL are intertwined
    global CRDS_REF_SUBDIR_MODE, _CRDS_CACHE_READONLY
    clear_crds_state()
    log.set_verbose(old_state["CRDS_VERBOSITY"])
    _CRDS_CACHE_READONLY = old_state.pop("_CRDS_CACHE_READONLY")
    CRDS_REF_SUBDIR_MODE = old_state["CRDS_REF_SUBDIR_MODE"]
    for key, val in old_state.items():
        os.environ[key] = str(val)
    if os.environ.get("CRDS_SERVER_URL"):
        api.set_crds_server(os.environ["CRDS_SERVER_URL"])
    if os.environ.get("CRDS_CWD"):
        os.chdir(os.environ["CRDS_CWD"])

def clear_crds_state():
    """Wipe out the existing configuration variable state of CRDS.
    """
    global CRDS_REF_SUBDIR_MODE, _CRDS_CACHE_READONLY
    for var in list(os.environ.keys()):
        if var.startswith("CRDS_"):
            del os.environ[var]
    CRDS_REF_SUBDIR_MODE = "None"
    _CRDS_CACHE_READONLY = False

# -------------------------------------------------------------------------------------
# Identifier used  to connect to remote status channels to monitor submission status, etc.

PROCESS_KEY_RE_STR = r"[a-zA-Z0-9:\.\-]{1,128}"   #  character "/" must be excluded
PROCESS_KEY_RE = re.compile(complete_re(PROCESS_KEY_RE_STR))

USER_NAME_RE_STR = r"[a-z_]+"
USER_NAME_RE = re.compile(complete_re(USER_NAME_RE_STR))

# -------------------------------------------------------------------------------------

# approximate regex to validate semver.org version numbers.

VERSION_RE_STR = complete_re(
    r"(\d{1,8}(\.\d{1,8}(\.\d{1,8})?)?)[a-zA-Z0-9_\+\-\.]{0,128}")
VERSION_RE = re.compile(VERSION_RE_STR)

def simplify_version(version):
    """
    >>> simplify_version('0.13.2a.dev185+gdb49aea9')
    '0.13.2'
    >>> simplify_version('0.7.8rc1.dev305')
    '0.7.8'
    >>> simplify_version('0.7.8')
    '0.7.8'
    >>> simplify_version('0.7')
    '0.7'
    >>> simplify_version('0')
    '0'
    >>> simplify_version('0.foo')
    '0'
    """
    return VERSION_RE.match(version).group(1)

# -------------------------------------------------------------------------------------

# These configure CRDS to stream mappings directly from S3 and return s3:// URLs to
# reference files.  It is expected that the files be stored on S3 in the "flat" structure
# (no instrument in the key).

S3_ENABLED = BooleanConfigItem("CRDS_S3_ENABLED", False, "When True, enables S3 streaming")

S3_RETURN_URI = BooleanConfigItem("CRDS_S3_RETURN_URI", False, "When True, getreferences() returns S3 URIs instead of cache paths.")

OBSERVATORY = StrConfigItem("CRDS_OBSERVATORY", None,
    "Configured observatory, required for S3 streaming",
    lower=True
)

# -------------------------------------------------------------------------------------
def get_uri(filename):
    """Return an appropriate s3:// or https:// URI for the specified filename based
    on optional CRDS config variables:

    CRDS_CONFIG_URI
    CRDS_MAPPING_URI
    CRDS_REFERENCE_URI
    CRDS_PICKLE_URI

    If the appropriate env var is not defined,  return "none".
    """
    if is_exception_condition(filename):
        return filename
    elif is_config(filename):
        uri = CRDS_CONFIG_URI.get()
    elif is_pickle(filename):
        uri = CRDS_PICKLE_URI.get()
    elif is_mapping(filename):
        uri = CRDS_MAPPING_URI.get()
    elif is_reference(filename):
        uri = CRDS_REFERENCE_URI.get()
    else:
        raise exceptions.CrdsError(f"Uknown file type for: '{filename}'")
    if uri == "none":
        return  uri
    else:
        if not uri.endswith("/"):
            uri += "/"
        uri += filename
        return uri

def is_exception_condition(filename):
    """Return True IFF `filename` describes some form of error or warning condition
    instead of a true filename.
    """
    return filename.upper().startswith("NOT FOUND")

# -------------------------------------------------------------------------------------

def test():
    """Run doctests on crds.config module."""
    import doctest
    from . import config
    return doctest.testmod(config)

if __name__ ==  "__main__":
    print(test())



"""This module provides a command line program to monitor a server side
process which communicates via client log messages.
"""

# ===================================================================
import sys
import time
import html

import numpy as np

from crds.core import exceptions, log, cmdline
from crds.core.log import srepr
from crds.client import api

# ===================================================================

_STATUS_COMPLETED = 0
_STATUS_FAILED = 1
_STATUS_CANCELLED = 3

class MonitorScript(cmdline.Script):
    """Command line script to dump real time submission progress messages."""

    description = """Command line script to dump real time submission progress messages.

"""

    epilog = """Monitoring is done with respect to a submission id/key and currently
polls the server for new messages at some periodic rate in seconds:

% crds monitor --poll-delay 3.0 --key 81323850-9517-416c-ae88-e6481de10a71
"""

    def __init__(self, *args, **keys):
        super(MonitorScript, self).__init__(*args, **keys)
        self._last_id = 0

        self.result = None

    def add_args(self):
        """Add class-specifc command line parameters."""
        super(MonitorScript, self).add_args()
        self.add_argument("--key", type=cmdline.process_key,
                          help="Key used to connect to remote process status stream.")
        self.add_argument("--poll-delay", type=int, default=3.0,
                          help="Time in seconds between polling for messages.")

    def main(self):
        """Main control flow of submission directory and request manifest creation."""
        log.divider("monitoring server on " + repr(self.args.key), char="=")
        exit_status = None
        while exit_status is None:
            for message in self._poll_status():
                handler = getattr(self, "handle_" + message.type, self.handle_unknown)
                exit_status = handler(message)
            time.sleep(self.args.poll_delay)
        log.divider("monitoring server done", char="=")
        return exit_status

    def _poll_status(self):
        """Use network API to pull status messages from server."""
        try:
            messages = api.jpoll_pull_messages(self.args.key, since_id=str(self._last_id))
            if messages:
                self._last_id = np.max([int(msg.id) for msg in messages])
            return messages
        except exceptions.StatusChannelNotFoundError:
            log.verbose("Channel", srepr(self.args.key),
                        "not found.  Waiting for processing to start.")
            return []
        except exceptions.ServiceError as exc:
            log.verbose("Unhandled RPC exception for", srepr(self.args.key), "is", str(exc))
            raise

    def format_remote(self, *params):
        """Format tuple of message `params` in a standardized way for messages
        coming from the remote process being monitored.
        """
        text = html.unescape(" ".join(params))
        return log.format(">>", text).strip()

    def handle_log_message(self, message):
        """Early API has only one message format,  "log_message".  Issue message info
        and continue monitoring.
        """
        log.info(self.format_remote(message.data))
        return None

    def handle_unknown(self,  message):
        """Handle unknown `message` types by issuing a warning and continuing monitoring."""
        log.warning(self.format_remote("Unknown message type", repr(message.type), "in", repr(message)))
        return None

    def handle_done(self, message):
        """Generic "done" handler issue info() message and stops monitoring / exits."""
        status = message.data["status"]
        result = message.data.get("result", None)

        if status == _STATUS_COMPLETED:
            log.info(self.format_remote("COMPLETED:", result))
        elif status == _STATUS_FAILED:
            log.error(self.format_remote("FAILED:", result))
        elif status == _STATUS_CANCELLED:
            log.error(self.format_remote("CANCELLED:", result))
        else:
            log.info(self.format_remote("DONE:", result))

        self.result = result

        return status

    def handle_cancel(self, message):
        """Generic "cancel" handler reports on commanded cancellation of remote process
        and possibly why it was cancelled.   Then stops monitoring /exits.
        """
        log.warning(self.format_remote("Processing cancelled:", message.data))

        self.result = message.data["result"]

        return _STATUS_CANCELLED

    def handle_fail(self, message):
        """Generic "fail" handler reports on remote process fatal error / failure
        and issues an error() message, then stops monitoring /exits.
        """
        log.error(self.format_remote("Processing failed:",  message.data))

        self.result = message.data["result"]

        return _STATUS_FAILED

    def handle_error(self, message):
        """Generic "error" handler issues an error message from remote process and
        continues monitoring.
        """
        log.error(self.format_remote(message.data))
        return None

    def handle_verbose(self, message):
        """Generic "verbose" handler issues a debug message from remote process if
        this monitor is running verbosely.
        """
        log.verbose(self.format_remote(message.data))
        return None

    def handle_warning(self, message):
        """Generic "warning" handler issues a  warning from remote process and
        contiues monitoring.
        """
        log.warning(self.format_remote(message.data))
        return None

# ===================================================================

if __name__ == "__main__":
    sys.exit(MonitorScript()())



"""This module is a command line script which handles comparing the best
reference recommendations for a particular context and dataset files.

For more details on the several modes of operations and command line parameters browse the source or run:

% crds bestrefs --help
"""
import sys
import os
from collections import namedtuple, OrderedDict

# ===================================================================

import crds
from crds.core import log, config, utils, timestamp, cmdline, heavy_client
from crds import diff, matches
from . import table_effects, headers
from crds.client import api

# ===================================================================

MIN_DATE = "1900-01-01 00:00:00"
MAX_DATE = "9999-01-01 23:59:59"

# ===================================================================

UpdateTuple = namedtuple("UpdateTuple", ["instrument", "filekind", "old_reference", "new_reference"])

# ============================================================================


def reformat_date_or_auto(date):
    """Add 'auto' as an extra valid value for --datasets-since dates.  Auto means figure out dates-since
    based on USEAFTER dates (recorded in rmaps as DATE-OBS TIME-OBS or META.OBSERVATION.DATE).
    """
    if date is None:
        return date
    elif date.lower() == "auto":
        return "auto"
    else:
        return timestamp.reformat_date(date)


class BestrefsScript(cmdline.Script, cmdline.UniqueErrorsMixin):
    """Command line script for determining best references for a sequence of dataset files."""

    description = """
The crds.bestrefs program runs the CRDS library to interpret CRDS reference
file assignment rules with respect to dataset parameters.

crds.bestrefs has several use cases which have different modes for fetching
input parameters, evaluating bestrefs and/or doing comparisons, and producing output.

crds.bestrefs runs in the HST archive pipeline to populate dataset headers FITS
keywords (e.g. DARKFILE) with best reference files.

Other modes of crds.bestrefs are used to support CRDS reprocessing or to test new
versions of CRDS rules.

The crds.bestrefs program is not normally used for JWST and best references
are assigned automatically as a consequence of running the CAL code.

* Determines best references with respect to a context or contexts.
* Optionally updates the headers of file-based data with new recommendations.
* Optionally compares new results to prior results.
* Optionally prints source data names affected by the new context.
    """

    epilog = """
.......................
crds.bestrefs use cases
.......................

  1. File (HST Pipeline) Mode

  The --files switch can be used to specify a list of FITS dataset files to
  process.  This is used in the HST pipeline in conjunction with
  --update-headers to fill in dataset FITS headers with recommended best
  references::

    % crds bestrefs  --update-headers  --files j8bt05njq_raw.fits ...

  The outcome of this command is updating the best references in the FITS
  headers of the specified .fits files.

  Omitting --update-headers can be used to evaluate bestrefs without altering
  the input FITS files::

    % crds bestrefs --print-new-references --files  j8bt05njq_raw.fits ...

  The --new-context switch can be used to choose a context which is not the
  current default::

    % crds bestrefs --new-context hst_0457.pmap --files ...

  2. Reprocessing Mode

  The bestrefs reprocessing mode is used in conjunction with archive databases
  to determine datasets which should be reprocessed as a consequence of the
  delivery of new reference files.

  Reprocessing mode evaluates the same dataset parameters with respect to an
  old context and a new context and recommends reprocessing datasets where some
  reference file assignment changes.

  Bestrefs reprocessing mode is run automatically on the CRDS servers whenever
  new reference files are delivered, after the new CRDS context is selected for
  use by the archive pipeline.  It is run e.g. like this::

    % crds bestrefs --affected-datasets --old-context  hst_0001.pmap --new-context hst_0002.pmap

  --affected-datasets is a "bundle switch" that captures standard options for
  reprocessing.  See *crds bestrefs --help* for more information on individual
  switches.

  Running reprocessing mode requires setting *CRDS_SERVER_URL*.

  3. Context Testing Mode

  CRDS bestrefs and the archive reprocessing parameters can also provide a
  quick way to evaluate a new context and/or residual errors.  It can answer
  the question "what classes of errors still exist for the latest context with
  respect to known parameter sets?"

  Context testing mode can be run like this::

    % crds bestrefs --check-context --new-context jwst-edit

  Context testing also requires setting *CRDS_SERVER_URL* to obtain archived
  dataset parameters.  Note that during JWST pre-I&T the archive database often
  contains parameter sets related to obsolete test cases.

  Undesired test cases can be weeded out like this::

    % crds bestrefs --check-context --new-context jwst-edit --drop-ids JW93135336001_02102_00001.MIRIFUSHORT:JW93135336001_02102_00001.MIRIFUSHORT

...........
New Context
...........

crds.bestrefs always computes best references with respect to a context which
can be explicitly specified with the --new-context parameter.  If --new-context
is not specified, the default operational context is determined by consulting
the CRDS server or looking in the local cache.

...........
Old Context
...........

--old-context can be used to specify a second context for which bestrefs
are dynamically computed; --old-context implies that a bestrefs comparison
will be made with --new-context.  If --old-context is not specified, it
defaults to None.  --old-context is only used for context-to-context
comparisons,  nominally for CRDS repro.

........................
Lookup Parameter Sources
........................

The following methods can be used to define parameter sets for which to compute
best references::

  --files can be used to specify a list of FITS files from which to load
    parameters and optionall update headers.

  --instruments can be used to specify a list of instruments.  Without
    --diffs-only or --datasets-since this choice selects ALL datasets for the
    specified instruments.

  --all-instruments is shorthand for all --instruments supported by the project.
    This parameter can be so memory intensive as to be infeasible.

  --datasets is used to specify a list of dataset IDs as would be found under --instruments.

  --load-pickles can be used to specify a list of .pkl or .json files that define parameter
    sets.  These can most easily be created using --save-pickle.

................
Comparison Modes
................

The --old-context and --compare-source-bestrefs parameters define the best
references comparison mode.  Each names the origin of a set of prior
recommendations and implicitly requests a comparison to the recommendations
from the newly computed bestrefs determined by --new-context.

*--old-context CONTEXT* specifies that the reference results should be
*computed* using the named context.

*--compare-source-bestrefs* directs that prior reference assignments should be
taken from the same *stored source* which provides matching parameters.  These
could be from FITS header keywords (e.g. DARKFILE), from live archive
parameters, or from prior parameter sets that have been stored in CRDS .json or
Python pickle files.

......................
Pickle and .json saves
......................

crds.bestrefs can load parameters and past results from a sequence of .pkl or
.json files using --load-pickles.  These are combined into a single parameter
source in command line order.

crds.bestrefs can save the parameters obtained from various sources into .pkl
or .json formatted save files using --save-pickle.  The single combined result
of multiple pickle or instrument parameter sources is saved.  The file
extension (.json or .pkl) defines the format used.

The preferred .json format defines a singleton { id: parameters}
dictionary on each line as a series of isolated .json objects.  Strictly
speaking only each individual line is .json,  but this localizes any errors.

.json format is preferred over .pkl because it is more transparent and robust
across different versions of Python.

.........
Verbosity
.........

crds.bestrefs has --verbose and --verbosity=N parameters which can increase the
amount of informational and debug output.  Verbosity ranges from 0..100 where 0
means "no debug output" and 100 means "all debug output".  50 is the default
for --verbose.

.........
Bad Files
.........

CRDS files can be designated as scientifically invalid on the CRDS server by
the CRDS team.  Knowledge of bad files is synchronized to remote caches by
crds.bestrefs and crds.sync.  By default, attempting to use bad rules or assign
bad references will generate errors and fail.  crds.bestrefs supports two
command line switches, *---allow-bad-rules* and *---allow-bad-references* to
override the default handling of bad files and enable their use with warnings.
Environment variables **CRDS_ALLOW_BAD_RULES** and
**CRDS_ALLOW_BAD_REFERENCES** can also be set to 1 to establish warnings rather
than errors as the default.
    """

    def __init__(self, *args, **keys):
        cmdline.Script.__init__(self, *args, **keys)

        # Placeholders until complex init is done.
        self.instruments = None
        self.oldctx = None
        self.newctx = None

        if self.args.regression:
            self.args.compare_source_bestrefs = True
            self.args.differences_are_errors = True
            self.args.stats = True
            self.args.dump_unique_errors = True

        if self.args.affected_datasets:
            self.args.diffs_only = True
            if not self.args.datasets_since:
                self.args.datasets_since = "auto"
            self.args.print_update_counts = True
            self.args.print_affected = True
            self.args.dump_unique_errors = True
            self.args.stats = True
            self.args.undefined_differences_matter = True
            self.args.na_differences_matter = True

        if self.args.check_context:
            self.args.dump_unique_errors = True
            self.args.stats = True
            self.args.undefined_differences_matter = True
            self.args.na_differences_matter = True
            if not self.args.all_errors_file:
                self.args.all_errors_file = "all_errors.ids"
            if not self.args.unique_errors_file:
                self.args.unique_errors_file = "unique_errors.ids"

        config.ALLOW_BAD_RULES.set(self.args.allow_bad_rules)
        config.ALLOW_BAD_REFERENCES.set(self.args.allow_bad_references)

        cmdline.UniqueErrorsMixin.__init__(self, *args, **keys)

        self.updates = OrderedDict()  # map of reference updates
        self.kill_list = OrderedDict()

        self.process_filekinds = [typ.lower() for typ in self.args.types ]    # list of filekind str's

        self.skip_filekinds = [typ.lower() for typ in self.args.skip_types]
        self.affected_instruments = None

        # See also complex_init()
        self.new_context = None     # Mapping filename
        self.old_context = None     # Mapping filename

        # headers corresponding to the new context
        self.new_headers = headers.HeaderGenerator(None,None,None)  # Abstract for pylint, replaced later

        # comparison variables
        self.compare_prior = None       # bool
        self.old_headers = headers.HeaderGenerator(None,None,None)  # Abstract for pylint, replaced later
        self.old_bestrefs_name = None   # info str identifying comparison results source,  .pmap filename or text

        self.pickle_headers = None  # any headers loaded from pickle files

        if self.args.remote_bestrefs:
            os.environ["CRDS_MODE"] = "remote"

        self.datasets_since = self.args.datasets_since

        self.active_header = None   # new or old header last processed with bestrefs
    def complex_init(self):
        """Complex init tasks run inside any --pdb environment,  also unfortunately --profile."""

        assert not (self.args.sync_references and self.readonly_cache), "Readonly cache,  cannot fetch references."

        self.new_context, self.old_context = self.setup_contexts()

        # Support 0 to 1 mutually exclusive source modes and/or any number of pickles
        exclusive_source_modes = [self.args.files, self.args.datasets, self.args.instruments,
                                  self.args.diffs_only, self.args.all_instruments]
        source_modes = len(exclusive_source_modes) - exclusive_source_modes.count(None)
        using_pickles = int(bool(self.args.load_pickles))
        assert source_modes <= 1 and (source_modes + using_pickles) >= 1, \
            "Must specify one of: --files, --datasets, --instruments, --all-instruments, --diffs-only and/or --load-pickles."

        if self.args.diffs_only:
            assert self.new_context and self.old_context, \
                "--diffs-only only works for context-to-context bestrefs."
            differ = diff.MappingDifferencer(
                self.observatory, self.old_context, self.new_context,
                include_header_diffs=True, hide_boring_diffs=True)
            self.affected_instruments = differ.get_affected()
            log.info("Mapping differences from", repr(self.old_context),
                     "-->", repr(self.new_context), "affect:\n",
                     log.PP(self.affected_instruments))
            self.instruments = self.affected_instruments.keys()
            if not self.instruments:
                log.info("No instruments were affected.")
                return False
            if (self.args.datasets_since == "auto" and
                    (differ.header_modified() or differ.files_deleted())):
                log.info("Checking all dates due to header changes or file deletions.")
                self.args.datasets_since = MIN_DATE
        elif self.args.instruments:
            self.instruments = self.args.instruments
        elif self.args.all_instruments:
            instruments = list(self.obs_pkg.INSTRUMENTS)
            for instr in ("all","system","synphot"):
                if instr in instruments:
                    instruments.remove(instr)
            self.instruments = instruments
        else:
            self.instruments = []

        if self.args.datasets_since == "auto":
            datasets_since = self.auto_datasets_since()
        else:
            datasets_since = self.args.datasets_since

        # headers corresponding to the new context
        self.new_headers = self.init_headers(self.new_context, datasets_since)

        self.compare_prior, self.old_headers, self.old_bestrefs_name = self.init_comparison(datasets_since)

        if not self.compare_prior:
            log.info("No comparison context or source comparison requested.")

        if self.args.files and not self.args.update_bestrefs:
            log.info("No file header updates requested;  dry run.  Use --update-bestrefs to update FITS headers.")

        return True

    @property
    def only_ids(self):
        return self._normalized(self.args.only_ids) or None

    @property
    def drop_ids(self):
        return self._normalized(self.args.drop_ids)

    def _normalized(self, id_list):
        if id_list:
            return [self.normalize_id(dataset) for dataset in id_list]
        else:
            return []

    def normalize_id(self, dataset):
        """Convert a given `dataset` ID to uppercase.  For the sake of simplicity convert
        simple IDs into unassociated exposure IDs in <exposure>:<exposure> form.  This is a
        convenience for JWST where currently the <product> term is always identical to
        <exposure>.  Where they're different as-in associated exposures for HST,  you must
        specify the drop ID fully to avoid misinterpretation as an unassociated exposure.
        """
        dataset = dataset.upper()
        if ":" not in dataset:
            dataset = dataset + ":" + dataset
        return dataset

    def auto_datasets_since(self):
        """Support --datasets-since="auto" and compute min EXPTIME for all references determined by diffs.

        Returns { instrument: EXPTIME, ... }
        """
        datasets_since = {}
        self.oldctx = crds.get_pickled_mapping(self.old_context)   # reviewed
        self.newctx = crds.get_pickled_mapping(self.new_context)   # reviewed
        for instrument in self.oldctx.selections:
            old_imap = self.oldctx.get_imap(instrument)
            new_imap = self.newctx.get_imap(instrument)
            added_references = diff.get_added_references(old_imap, new_imap)
            deleted_references = diff.get_deleted_references(old_imap, new_imap)
            added_exp_time = deleted_exp_time = MAX_DATE
            if added_references:
                added_exp_time = matches.get_minimum_exptime(new_imap, added_references)
            if deleted_references:
                deleted_exp_time = matches.get_minimum_exptime(old_imap, deleted_references)
            exp_time = min(added_exp_time, deleted_exp_time)
            if exp_time != MAX_DATE:  # if a USEAFTER min found,  remember it.
                datasets_since[instrument] = exp_time
        log.info("Possibly affected --datasets-since dates determined by",
                 repr(self.old_context), "-->", repr(self.new_context), "are:\n", log.PP(datasets_since))
        return datasets_since

    def add_args(self):
        """Add bestrefs script-specific command line parameters."""

        self.add_argument("-n", "--new-context", dest="new_context",
                          help="Compute the updated best references using this context. "
                          "Uses current operational context by default.",
                          default=None, type=cmdline.mapping_spec)

        self.add_argument("-o", "--old-context", dest="old_context",
                          help="Compare bestrefs recommendations from two contexts.",
                          metavar="OLD_CONTEXT", default=None, type=cmdline.mapping_spec)

        self.add_argument("--fetch-old-headers", dest="fetch_old_headers", action="store_true",
                          help="Fetch old headers in accord with old parameter lists.   Slower,  avoid unless required.")

        self.add_argument("-f", "--files", nargs="+", metavar="FILES", default=None,
                          help="Dataset files to compute best references for and optionally update headers.")

        self.add_argument("-d", "--datasets", nargs="+", metavar="IDs", default=None,
                          help="Dataset ids to consult database for matching parameters and old results.")

        self.add_argument("--all-instruments", action="store_true", default=None,
                          help="Compute best references for cataloged datasets for all supported instruments in database.")

        self.add_argument("-i", "--instruments", nargs="+", metavar="INSTRUMENTS", default=None,
                          help="Instruments to compute best references for, all historical datasets in database.")

        self.add_argument("-p", "--load-pickles", nargs="*", default=None,
                          help="Load dataset headers and prior bestrefs from pickle files,  in worst-to-best update order.  Can also load .json files.")

        self.add_argument("-a", "--save-pickle", default=None,
                          help="Write out the combined dataset headers to the specified pickle file.  Can also store .json file.")

        self.add_argument("-t", "--types", nargs="+",  metavar="REFERENCE_TYPES",  default=(),
                          help="Explicitly define the list of reference types to process, --skip-types also still applies.")

        self.add_argument("-k", "--skip-types", nargs="+",  metavar="SKIPPED_REFERENCE_TYPES",  default=(),
                          help="A list of reference types which should not be processed,  defaulting to nothing.")

        self.add_argument("--all-types", action="store_true",
                          help="Evaluate every reference file type regardless of dataset exposure type.")

        self.add_argument("--diffs-only", action="store_true", default=None,
                          help="For context-to-context comparison, choose only instruments and types from context differences.")

        self.add_argument("--datasets-since", default=None, type=reformat_date_or_auto,
                          help="Cut-off date for datasets, none earlier than this.  Use 'auto' to exploit reference USEAFTER.  OFF by default.")

        self.add_argument("-c", "--compare-source-bestrefs", dest="compare_source_bestrefs", action="store_true",
                          help="Compare new bestrefs recommendations to recommendations from data source,  files or database.")

        self.add_argument("--update-pickle", action="store_true",
                          help="Replace source bestrefs with CRDS bestrefs in output pickle.  For setting up regression tests.")

        self.add_argument("--only-ids", nargs="*", default=None, dest="only_ids", metavar="IDS",
                          help="If specified, process only the listed dataset ids.")

        self.add_argument("--drop-ids", nargs="*", default=[], dest="drop_ids", metavar="IDS",
                          help="If specified, skip these dataset ids.")

        self.add_argument("-u", "--update-bestrefs",  dest="update_bestrefs", action="store_true",
                          help="Update sources with new best reference recommendations.")

        self.add_argument("--print-affected", dest="print_affected", action="store_true",
                          help="Print names of products for which the new context would assign new references for some exposure.")

        self.add_argument("--print-affected-details", action="store_true",
                          help="Include instrument and affected types in addition to compound names of affected exposures.")

        self.add_argument("--print-new-references", action="store_true",
                          help="Prints one line per reference file change.  If no comparison requested,  prints all bestrefs.")

        self.add_argument("--print-update-counts", action="store_true",
                          help="Prints dictionary of update counts by instrument and type,  status on updated files.")

        self.add_argument("--print-error-headers", action="store_true",
                          help="For each tracked error,  print out the corresponding dataset header for offline analysis.")

        self.add_argument("-r", "--remote-bestrefs", action="store_true",
                          help="Compute best references on CRDS server,  convenience for env var CRDS_MODE='remote'")

        self.add_argument("-m", "--sync-mappings", default="1", dest="sync_mappings", type=int,
                          help="Fetch the required context mappings to the local cache.  Defaults TRUE.")

        self.add_argument("-s", "--sync-references", default="0", dest="sync_references", type=int,
                          help="Fetch the refefences recommended by new context to the local cache. Defaults FALSE.")

        self.add_argument("--differences-are-errors", action="store_true",
                          help="Treat recommendation differences between new context and original source as errors.")

        self.add_argument("--allow-bad-rules", action="store_true",
                          help="Only warn if a context which is marked 'bad' is used, otherwise error.")

        self.add_argument("--allow-bad-references", action="store_true",
                          help="Only warn if a reference which is marked bad is recommended, otherwise error.")

        self.add_argument("--undefined-differences-matter", action="store_true",
                          help="If not set, a transition from UNDEFINED to anything else is not considered a difference error.")

        self.add_argument("--na-differences-matter", action="store_true",
                          help="If not set,  either CDBS or CRDS recommending N/A is OK to mismatch.")

        self.add_argument("-g", "--regression", action="store_true",
                          help="Abbreviation for --compare-source-bestrefs --differences-are-errors --dump-unique-errors --stats")

        self.add_argument("--check-context", action="store_true",
                          help="Abbreviation for --undefined-differences-matter "
                          "--na-differences-matter --dump-unique-errors --stats")

        self.add_argument("--affected-datasets", action="store_true",
                          help="Abbreviation for --diffs-only --datasets-since=auto --undefined-differences-matter "
                          "--na-differences-matter --print-update-counts --print-affected --dump-unique-errors --stats")

        self.add_argument("-z", "--optimize-tables", action="store_true",
                          help="If set, apply row-based optimizations to screen out inconsequential table updates.")

        self.add_argument("--eliminate-duplicate-cases", action="store_true",
                          help="Categorize unique bestrefs results as errors to determine representative test cases...  Replaces normal error counts with coverage counts and ids.")

        cmdline.UniqueErrorsMixin.add_args(self)

    def setup_contexts(self):
        """Determine and cache the new and comparison .pmap's for this run."""
        if self.args.new_context is None:
            log.verbose("Using default new context", repr(self.default_context),
                        "for computing updated best references.", verbosity=25)
            new_context = self.default_context
        else:
            log.verbose("Using explicit new context", repr(self.args.new_context),
                        "for computing updated best references.", verbosity=25)
            new_context = self.resolve_context(self.args.new_context)
        if self.args.old_context is not None:
            log.verbose("Using explicit old context", repr(self.args.old_context), verbosity=25)
            old_context = self.resolve_context(self.args.old_context)
        else:
            old_context = None
        if self.server_info.effective_mode != "remote":
            if old_context is not None and not os.path.dirname(old_context):
                self.dump_mappings([old_context])
            if not os.path.dirname(new_context):
                self.dump_mappings([new_context])
        return new_context, old_context

    @utils.cached
    def warn_bad_context(self, name, context, instrument):
        """Issue a warning if `context` of named `name` is a known bad file."""
        # Get subset of bad files contained by this context.
        if context is None:
            return
        bad_contained = heavy_client.get_bad_mappings_in_context(self.observatory, context, instrument)
        if bad_contained:
            if not config.ALLOW_BAD_RULES:
                self.log_and_track_error("ALL", "ALL", "ALL", name, "=", repr(context),
                                         "is bad or contains bad rules.  Use is not recommended,  results may not be scientifically valid.")
            else:
                log.warning(name, "=", repr(context),
                            "is bad or contains bad rules.  Use is not recommended,  results may not be scientifically valid.")
            log.verbose(name, "=", repr(context), "contains bad rules", repr(bad_contained))

    def warn_bad_reference(self, dataset, instrument, filekind, reference):
        """Issue a warning if `reference` is a known bad file."""
        if reference.lower() in self.bad_files:
            if not config.ALLOW_BAD_REFERENCES:
                self.log_and_track_error(dataset, instrument, filekind, "File", repr(reference),
                                         "is bad. Use is not recommended,  results may not be scientifically valid.")
            else:
                log.warning("For", dataset, instrument, filekind, "File", repr(reference),
                            "is bad. Use is not recommended,  results may not be scientifically valid.")
            return 1
        else:
            return 0

    def warn_bad_updates(self):
        """Issue warnings for each bad references in the updates map.  This is not inlined in the
        getreferences() core function because it only pertains to *new* bad reference assignments,
        not old ones,  so it is only relevant for an update,  not every assignment.
        """
        log.verbose("Checking updates for bad files.")
        bad_files = 0
        for (dataset, updates) in sorted(self.updates.items()):
            for update in sorted(updates):
                bad_files += self.warn_bad_reference(dataset, update.instrument, update.filekind, update.new_reference)
        log.verbose("Total bad files =", bad_files)

    def locate_file(self, filename):
        """Locate a dataset file leaving the path unchanged. Applies to self.args.files"""
        return filename

    def init_headers(self, context, datasets_since):
        """Create header a header generator for `context`,  interpreting command line parameters."""
        if self.args.files:
            the_headers = headers.FileHeaderGenerator(context, self.files, datasets_since)
            # log.info("Computing bestrefs for dataset files", self.args.files)
        elif self.args.datasets:
            self.require_server_connection()
            the_headers = headers.DatasetHeaderGenerator(context, [dset.upper() for dset in self.args.datasets], datasets_since)
            log.info("Computing bestrefs for datasets", repr(self.args.datasets))
        elif self.instruments:
            self.require_server_connection()
            log.info("Computing bestrefs for db datasets for", repr(list(self.instruments)))
            the_headers = headers.InstrumentHeaderGenerator(
                context, self.instruments, datasets_since, self.args.save_pickle, self.server_info)
        elif self.args.load_pickles:
            the_headers = None
        else:
            log.error("Invalid header source configuration.   "
                      "Specify --files, --datasets, --instruments, --all-instruments, or --load-pickles.")
            self.print_help()
            sys.exit(-1)
        if self.args.load_pickles:
            self.pickle_headers = headers.PickleHeaderGenerator(
                context, self.args.load_pickles, only_ids=self.only_ids, datasets_since=datasets_since)
            if the_headers:   # combine partial correction headers field-by-field
                log.verbose("Augmenting primary parameter sets with pickle overrides.")
                the_headers.update_headers(self.pickle_headers.headers, only_ids=self.only_ids)
            else:   # assume pickles-only sources are all complete snapshots
                log.verbose("Computing bestrefs solely from pickle files:", repr(self.args.load_pickles))
                the_headers = self.pickle_headers
        return the_headers

    def init_comparison(self, datasets_since):
        """Interpret command line parameters to determine comparison mode."""
        assert not (self.args.old_context and self.args.compare_source_bestrefs), \
            "Cannot specify both --old-context and --compare-source-bestrefs."
        compare_prior = \
            self.args.old_context or \
            self.args.compare_source_bestrefs or \
            self.args.print_affected or \
            self.args.print_affected_details
        # self.args.update_bestrefs or \
        if compare_prior:
            if self.args.old_context:
                old_fname = self.args.old_context
                if self.args.fetch_old_headers:
                    log.verbose("Fetching old headers independently allowing them to differ from new headers.")
                    old_headers = self.init_headers(self.old_context, datasets_since)
                else:
                    log.verbose_warning("Assuming parameter names and required types are the same across contexts.")
                    old_headers = self.new_headers
            else:
                old_fname = "recorded bestrefs"
                old_headers = self.new_headers
        else:
            old_headers = old_fname = None
        return compare_prior, old_headers, old_fname

    def main(self):
        """Compute bestrefs for datasets."""
        # Finish __init__() inside --pdb
        if self.complex_init():
            for i, dataset in enumerate(self.new_headers):
                if i != 0 and i % 1000 == 0:
                    log.verbose(self.get_stat("datasets"), "sources processed", verbosity=5)
                self.process(dataset)
            self.post_processing()
        self.report_stats()
        if self.args.eliminate_duplicate_cases:
            log.warning("Running in --eliminate-duplicate-cases mode;  even successful bestrefs are categorized as errors for analysis.")
        log.verbose(self.get_stat("datasets"), "sources processed", verbosity=5)
        log.verbose(len(self.unkilled_updates), "source updates", verbosity=5)
        log.standard_status()
        return log.errors()

    def process(self, dataset):
        """Process best references for `dataset`,  printing dataset output,  collecting stats, trapping exceptions."""
        with log.error_on_exception("Failed processing", repr(dataset)):
            log.verbose("=" * 120)
            if dataset in self.drop_ids:
                log.verbose("Skipping drop-list dataset", repr(dataset))
                return
            elif self.only_ids and dataset not in self.only_ids:
                log.verbose("Skipping", repr(dataset), "not in --only-ids", verbosity=80)
                return
            elif self.args.files:
                log.info("===> Processing", dataset)     # file mode
            else:
                log.verbose("===> Processing", dataset, verbosity=25)   # database or regression modes
            self.increment_stat("datasets", 1)
            self._process(dataset)

    def _process(self, dataset):
        """Core best references,  add to update tuples."""
        self.active_header = new_header = self.new_headers.get_lookup_parameters(dataset)
        instrument = utils.header_to_instrument(new_header)
        self.warn_bad_context("New-context", self.new_context, instrument)
        new_bestrefs = self.get_bestrefs(instrument, dataset, self.new_context, new_header)
        if self.compare_prior:
            self.warn_bad_context("Old-context", self.old_context, instrument)
            if self.args.old_context:
                self.active_header = old_header = self.old_headers.get_lookup_parameters(dataset)
                old_bestrefs = self.get_bestrefs(instrument, dataset, self.old_context, old_header)
            else:
                old_bestrefs = self.old_headers.get_old_bestrefs(dataset)
            updates, kill_list = self._compare_bestrefs(instrument, dataset, old_bestrefs, new_bestrefs)
            if self.args.optimize_tables:
                updates = self.optimize_tables(dataset, updates)
        else:
            updates, kill_list = self._screen_bestrefs(instrument, dataset, new_bestrefs)
        if self.args.update_pickle:  # XX  mutating input bestrefs to support updated pickles
            self.new_headers.update_headers({dataset: new_bestrefs})
        if updates:
            self.updates[dataset] = updates
        if kill_list:
            self.kill_list[dataset] = kill_list

    def get_bestrefs(self, instrument, dataset, context, header):
        """Compute the bestrefs for `dataset` with respect to loaded mapping/context `ctx`."""
        with log.augment_exception("Failed determining reference types for", repr(dataset),
                                   "with respect to", (instrument, context, header)):
            reftypes = self.determine_reftypes(instrument, dataset, context, header)
            if reftypes is None:
                return {}
        with log.augment_exception("Failed computing bestrefs for data", repr(dataset),
                                   "with respect to", repr(context)):
            bestrefs = crds.getrecommendations(
                header, reftypes=reftypes, context=context, observatory=self.observatory, fast=log.get_verbose() < 50)
        return {key.upper(): value for (key, value) in bestrefs.items()}

    def determine_reftypes(self, instrument, dataset, context, header):
        """Based on instrument, context, header as well as command line parameters determine the list
        of reftypes that should be processed.

        Incorporate knowledge from:

        1. --diffs-only   (context-to-context differences determine affected types)
        2. --types        explicit command line specification
        3. Applicable types from project specific locate.header_to_reftypes()
        4. --all-types    override 1-3 above,  check all
        5. --skip-types   explicit command line removal, reduce 1-4 above
        """
        applicable_types = set()
        with log.verbose_warning_on_exception("Failed determining reftypes for", repr(dataset)):
            applicable_types = set(self.locator.header_to_reftypes(header, context))
        if self.affected_instruments:
            types = set(self.affected_instruments[instrument.lower()])
            if applicable_types:
                types &= applicable_types
            if not types:
                return None
        elif self.args.types:
            types = set(self.process_filekinds)
        else:
            types = applicable_types
        if self.args.all_types or not types:  # --all-types trumps --types, --diffs-only
            pmap = crds.get_pickled_mapping(context)
            types = pmap.get_imap(instrument).selections.keys()
        if types and self.args.skip_types:
            types = set(types) - set(self.args.skip_types)
        types = sorted(list(types))
        return types

    @property
    def update_promise(self):
        """Return a string identifying that and update would or will occurr, depending on --update-bestrefs."""
        if self.args.update_bestrefs:
            return ":: Updating."
        else:
            return ":: Would update."

    no_update = ":: No update."

    def verbose_with_prefix(self, dataset, instrument, filekind, *args, **keys):
        """Output a verbose log message with a prefix formated using format_prefix()."""
        if log.should_output(*args, **keys):
            log.verbose(self.format_prefix(dataset, instrument, filekind), *args, **keys)

    def screen_bestrefs(self, instrument, dataset, newrefs):
        """Return only the update tuples of _screen_bestrefs."""
        return self._screen_bestrefs(instrument, dataset, newrefs)[0]

    def _screen_bestrefs(self, instrument, dataset, newrefs):
        """Scan best references dict `newrefs` for atypical results and issue errors and warnings.

        Returns ([UpdateTuple(), ...], [kill list, ...])
        """
        # XX  This is closely related to compare_bestrefs, maintain both!!   See also update_bestrefs()

        log.verbose("-" * 120, verbosity=55)

        updates, kill_list = [], []

        for filekind in sorted(newrefs):

            filekind = filekind.lower()

            if filekind in self.skip_filekinds:
                self.verbose_with_prefix(dataset, instrument, filekind, "Skipping type.", verbosity=55)
                continue

            new_ok, new = self.handle_na_and_not_found("New:", newrefs, dataset, instrument, filekind)
            update = UpdateTuple(instrument, filekind, None, new)
            if new_ok or self.args.update_pickle or self.args.eliminate_duplicate_cases:
                self.verbose_with_prefix(dataset, instrument, filekind,
                    "Bestref FOUND:", repr(new).lower(),  self.update_promise, verbosity=30)
                updates.append(update)
            else:  # ERROR's cannot update
                kill_list.append(update)

        return updates, kill_list

    def compare_bestrefs(self, instrument, dataset, oldrefs, newrefs):
        """Return only the update tuples of _compare_bestrefs()."""
        return self._compare_bestrefs(instrument, dataset, oldrefs, newrefs)[0]

    def _compare_bestrefs(self, instrument, dataset, oldrefs, newrefs):
        """Compare best references dicts `newrefs` and `oldrefs` for `instrument` and `dataset`.

        Returns ([UpdateTuple(), ...], [Kill list...])
        """
        # XX  This is closely related to screen_bestrefs,  maintain both!!    See also update_bestrefs()

        log.verbose("-" * 120, verbosity=55)

        updates, kill_list = [], []

        filekinds = sorted(list(set(list(newrefs.keys()) + list(oldrefs.keys()))))
        for filekind in filekinds:

            filekind = filekind.lower()

            if filekind in self.skip_filekinds:
                self.verbose_with_prefix(dataset, instrument, filekind, "Skipping type.", verbosity=55)
                continue

            _old_ok, old = self.handle_na_and_not_found("Old:", oldrefs, dataset, instrument, filekind)

            new_ok, new = self.handle_na_and_not_found("New:", newrefs, dataset, instrument, filekind)

            update = UpdateTuple(instrument, filekind, old, new)

            if not (new_ok or self.args.update_pickle):   # ERROR's in new cannot update,  except during regression capture
                kill_list.append(update)
                continue

            if new != old:
                if self.args.differences_are_errors:  # regression mode vs. reprocessing
                    #  By default, either CDBS or CRDS scoring a reference as N/A short circuits mismatch errors.
                    if (old != "N/A" and new != "N/A") or self.args.na_differences_matter:
                        self.log_and_track_error(dataset, instrument, filekind,
                            "Comparison difference:", sreprlow(old), "-->", sreprlow(new), self.update_promise)
                elif self.args.print_new_references or log.get_verbose() >= 30 or self.args.files:
                    log.info(self.format_prefix(dataset, instrument, filekind),
                             "New best reference:", sreprlow(old), "-->", sreprlow(new), self.update_promise)
                updates.append(update)
            else:
                self.verbose_with_prefix(dataset, instrument, filekind,
                    "Lookup MATCHES:", sreprlow(old), self.no_update,  verbosity=30)
        return updates, kill_list

    def handle_na_and_not_found(self, name, bestrefs, dataset, instrument, filekind):
        """Fetch the bestref for `filekind` from `bestrefs`, and handle conversions to N/A
        and CRDS NOT FOUND errors.

        `name` is a string identifier for this conversion, Old or New.
        `bestrefs` is a dictionary { filekind : bestref, }

        Return (ref_ok, raw_ref, ref)  where:
            raw_ref is the original name of the reference,  stripped of any iref$ prefix.
            ref is the fully normalized name of the reference, converted to N/A as needed.
            ref_ok is True IFF bestrefs did not fail altogether.
        """
        # UNDEFINED corresponds to "no value in database" or "type not defined in CRDS"
        ref_org = cleanpath(bestrefs.get(filekind.upper(), "UNDEFINED")).strip()
        ref = ref_org.upper()
        ref_ok = True
        if ref.startswith("NOT FOUND N/A"):
            self.verbose_with_prefix(dataset, instrument, filekind, "Bestref is natural N/A.", verbosity=60)
            ref = "N/A"
        elif ref in ("NOT FOUND NO MATCH FOUND.", "UNDEFINED", "NONE", "", "*"):
            if self.args.undefined_differences_matter:  # track these when N/A is being scrutinized, regardless of diff.
                self.log_and_track_error(
                    dataset, instrument, filekind, name,  "No match found:", repr(ref_org))
                ref_ok = False
            else:
                self.verbose_with_prefix(dataset, instrument, filekind,
                    name, "No match found:", repr(ref_org), " => 'N/A'.")
                ref = "N/A"
        elif ref.startswith("NOT FOUND"):
            self.log_and_track_error(
                dataset, instrument, filekind, name, "Bestref FAILED:", ref_org[len("NOT FOUND"):])
            ref_ok = False
        return ref_ok, ref

    def log_and_track_error(self, dataset, *pars, **keys):
        """Track and categorize the specified error,  printing out the dataset header
        if requested on the command line.
        """
        parts = dataset.split(":")
        if parts[0] == parts[-1]:  # no guarantee len() == 2
            dataset = parts[0]
        super(BestrefsScript, self).log_and_track_error(dataset, *pars, **keys)
        if self.args.print_error_headers:
            log.info("Header for", repr(dataset) + ":\n", log.PP(self.active_header))

    def post_processing(self):
        """Given the computed update list, print out results,  update file headers, and fetch missing references."""

        if self.args.save_pickle:
            self.new_headers.save_pickle(self.args.save_pickle, only_ids=self.only_ids)

        self.warn_bad_updates()  # Warn about bad file reference updates only, not failures

        if self.args.print_new_references:
            self.print_new_references()

        if self.args.update_bestrefs:
            log.verbose("Performing best references updates.")
            self.new_headers.handle_updates(self.updates)

        if self.args.sync_references:
            self.sync_references()

        if self.args.print_update_counts:
            self.print_update_stats()  # For affected datasets

        if self.args.print_affected:
            self.print_affected()

        if self.args.print_affected_details:
            self.print_affected_details()

        if self.args.eliminate_duplicate_cases:
            self.eliminate_duplicate_cases()

        self.dump_unique_errors()

    def optimize_tables(self, dataset, updates):
        """Drop table updates for which the reference change doesn't matter based upon examining the
        selected rows.
        """
        for update in sorted(updates):
            new_header = self.new_headers.header(dataset)
            if not table_effects.is_reprocessing_required(dataset, new_header, self.old_context, self.new_context, update):
                updates.remove(update)  # reprocessing not required, ignore update.
                log.verbose("Removing table update for", update.instrument, update.filekind, dataset,
                            "no effective change from reference", repr(update.old_reference),
                            "-->", repr(update.new_reference), verbosity=25)
        return updates

    @property
    def unkilled_updates(self):
        """Return only members of self.updates for which there is no corresponding kill list of failed bestrefs."""
        return { dataset:updates for (dataset, updates) in self.updates.items()
                 if updates and dataset not in self.kill_list }

    def print_affected(self):
        """Print the product id for any product which has new bestrefs for any
        of its component exposures.   All components share a common product id.
        """
        affected_products = { self.dataset_to_product_id(dataset)
                              for dataset in self.unkilled_updates }
        log.info("Affected products =", len(affected_products))
        for product in sorted(affected_products):
            print(product)
        sys.stdout.flush()

    def dataset_to_product_id(self, dataset):
        """CRDS manages products and associations using : separated compound IDs of indeterminate
        complexity.  The only thing guaranteed is that the first colon-section of the dataset ID is the
        product (for possible reprocessing) which is reported whenever any bestref changes for any
        dataset ID beginning with that prefix.
        """
        return dataset.split(":")[0].lower()

    def print_affected_details(self):
        """Print compound ID, instrument, and affected reference types for every exposure with new best references,
        one line per exposure.
        """
        for dataset in sorted(self.updates):
            if self.updates[dataset]:
                types = sorted([update.filekind for update in self.updates[dataset]])
                print("{} {} {}".format(dataset.lower(), self.updates[dataset][0].instrument.lower(), " ".join(types)))
        sys.stdout.flush()

    def print_update_stats(self):
        """Print update counts for each instrument and type."""
        stats = dict()
        for dataset in self.updates:
            for update in self.updates[dataset]:
                if update.instrument not in stats:
                    stats[update.instrument] = dict()
                if update.filekind not in stats[update.instrument]:
                    stats[update.instrument][update.filekind] = 0
                stats[update.instrument][update.filekind] += 1
        log.info("Updated exposure counts:\n", log.PP(stats))

    def print_new_references(self):
        """Print the compound id and update tuple for each exposure with updates."""
        for dataset in sorted(self.updates):
            for update in self.updates[dataset]:
                print(dataset.lower() + " " + " ".join([str(val).lower() for val in update]))
        sys.stdout.flush()

    def eliminate_duplicate_cases(self):
        """Special mode to categorize *single context* bestrefs results, good or
        errors, as unique error strings.
        """
        log.reset()
        self.clear_error_counts()
        for dataset in self.updates:
            updates = self.updates[dataset]
            self.log_and_track_error(
                dataset, "any", "any", "COVERAGE:", self.updates_repr(updates))

    def updates_repr(self, updates):
        """Convert a single dataset's list of update tuples into a string which
        can be used to define unique CRDS results for coverage purposes.
        """
        return str(sorted([tuple(update) for update in updates])).lower()

    def sync_references(self):
        """Locally cache the new references referred to by updates."""
        synced_references = {
            tup.new_reference.lower()
            for dataset in self.updates
            for tup in self.updates[dataset] if tup.new_reference not in ["N/A"]
        }
        api.dump_references(self.new_context, sorted(synced_references), raise_exceptions=self.args.pdb)

# ============================================================================

def sreprlow(s):
    """Squash unicode and return the repr() of string `s` as lower case."""
    return repr(str(s)).lower()

def cleanpath(name):
    """jref$n4e12510j_crr.fits  --> n4e12510j_crr.fits"""
    if "ref$" in name:
        return name.split("$")[-1].strip()
    elif name.startswith("crds://"):
        return name[len("crds://"):]
    else:
        return name

# ============================================================================

def assign_bestrefs(filepaths, context=None, reftypes=(),
                    sync_references=False, verbosity=-1):
    """Assign best references to FITS files specified by `filepaths`
    filling in appropriate reference type keywords.

    Define best references using either .pmap `context` or the default
    CRDS operational context if context=None.

    If `reftypes` is defined, assign bestrefs to only the listed
    reftypes, otherwise assign all reftypes.

    If `sync_references` is True, download any missing reference files
    to the CRDS cache.

    Verbosity defines the level of CRDS log output:

    verbosity=-3    feeling lucky, no output
    verbosity=-2    only errors
    verbosity=-1    only warnings and errors
    verbosity=0     info, warnings, and errors
    verbosity=10    info + minimal progress output
    verbosity=30    info + simplified bestref assignments
    verbosity=50    info + keywords + exact values (standard)
    verbosity=60    info + bestrefs elimination
    ...
    -3 <= verbosity <= 100

    NOTE: While assign_bestrefs() may work for JWST, it is primarily intended
    to expose the behavior of the program used in the HST archive pipeline for
    programmatic use.  Since JWST CAL normally uses an alternate mechanism for
    updating FITS files, using assign_bestrefs() for JWST is less applicable
    and more experimental.

    Returns  count of errors
    """
    old_verbosity = log.set_verbose(verbosity)

    cmd = "crds.bestrefs --update-bestrefs"

    if context:
        cmd += f" --new-context {context}"

    if reftypes:
        types = " ".join(reftypes)
        cmd += f" --types {types}"

    if sync_references:
        cmd += " --sync-references=1"

    files = " ".join(filepaths)
    cmd += f" --files {files}"

    script = BestrefsScript(cmd)

    errors = script()

    log.set_verbose(old_verbosity)

    return errors



"""This module contains doctests and unit tests which exercise the crds.matches program."""
import os, os.path
from pprint import pprint as pp

import crds
from crds.core import rmap, log, exceptions, config
from crds import tests
from crds.matches import MatchesScript
from crds.client import api
from crds.tests import test_config

from nose.tools import assert_raises, assert_true

# ==================================================================================

HERE = os.path.dirname(__file__) or "."

# ==================================================================================

def dt_matches_files():
    """
    >>> old_state = test_config.setup()

    >>> MatchesScript("crds.matches  --contexts hst_0001.pmap --files lc41311jj_pfl.fits")()
     lc41311jj_pfl.fits : ACS PFLTFILE DETECTOR='WFC' CCDAMP='A|ABCD|AC|AD|B|BC|BD|C|D' FILTER1='F625W' FILTER2='POL0V' OBSTYPE='IMAGING' FW1OFFST='N/A' FW2OFFST='N/A' FWSOFFST='N/A' DATE-OBS='1997-01-01' TIME-OBS='00:00:00'
    0
    >>> config.set_crds_state(old_state)
    """

def dt_matches_files_omit_parameters_brief():
    """
    >>> old_state = test_config.setup()
    >>> MatchesScript("crds.matches  --contexts hst_0001.pmap --files lc41311jj_pfl.fits --omit-parameter-names --brief-paths")()
     lc41311jj_pfl.fits :  'WFC' 'A|ABCD|AC|AD|B|BC|BD|C|D' 'F625W' 'POL0V' 'IMAGING' 'N/A' 'N/A' 'N/A' '1997-01-01' '00:00:00'
    0
    >>> config.set_crds_state(old_state)
    """

def dt_matches_tuple_format():
    """
    >>> old_state = test_config.setup()
    >>> MatchesScript("crds.matches --contexts hst.pmap --files lc41311jj_pfl.fits --tuple-format")()
     lc41311jj_pfl.fits : (('OBSERVATORY', 'HST'), ('INSTRUMENT', 'ACS'), ('FILEKIND', 'PFLTFILE'), ('DETECTOR', 'WFC'), ('CCDAMP', 'A|ABCD|AC|AD|B|BC|BD|C|D'), ('FILTER1', 'F625W'), ('FILTER2', 'POL0V'), ('OBSTYPE', 'IMAGING'), ('FW1OFFST', 'N/A'), ('FW2OFFST', 'N/A'), ('FWSOFFST', 'N/A'), ('DATE-OBS', '1997-01-01'), ('TIME-OBS', '00:00:00'))
    0
    >>> config.set_crds_state(old_state)
    """

def dt_matches_datasets_minimize_headers_contexts_condition():
    """
    >>> old_state = test_config.setup()
    >>> MatchesScript("crds.matches --datasets JBANJOF3Q --minimize-headers --contexts hst_0048.pmap hst_0044.pmap --condition-values")()
    JBANJOF3Q:JBANJOF3Q : hst_0044.pmap : APERTURE='WFC1-2K' ATODCORR='OMIT' BIASCORR='OMIT' CCDAMP='B' CCDCHIP='1.0' CCDGAIN='2.0' CRCORR='OMIT' DARKCORR='OMIT' DATE-OBS='2010-01-31' DETECTOR='WFC' DQICORR='PERFORM' DRIZCORR='OMIT' FILTER1='F502N' FILTER2='F660N' FLASHCUR='OFF' FLATCORR='OMIT' FLSHCORR='OMIT' FW1OFFST='0.0' FW2OFFST='0.0' FWSOFFST='0.0' GLINCORR='UNDEFINED' INSTRUME='ACS' LTV1='-2048.0' LTV2='-1.0' NUMCOLS='2070.0' NUMROWS='2046.0' OBSTYPE='INTERNAL' PCTECORR='OMIT' PHOTCORR='OMIT' REFTYPE='UNDEFINED' SHADCORR='OMIT' SHUTRPOS='B' TIME-OBS='01:07:14.960000' XCORNER='1.0' YCORNER='2072.0'
    JBANJOF3Q:JBANJOF3Q : hst_0048.pmap : APERTURE='WFC1-2K' ATODCORR='OMIT' BIASCORR='OMIT' CCDAMP='B' CCDCHIP='1.0' CCDGAIN='2.0' CRCORR='OMIT' DARKCORR='OMIT' DATE-OBS='2010-01-31' DETECTOR='WFC' DQICORR='PERFORM' DRIZCORR='OMIT' FILTER1='F502N' FILTER2='F660N' FLASHCUR='OFF' FLATCORR='OMIT' FLSHCORR='OMIT' FW1OFFST='0.0' FW2OFFST='0.0' FWSOFFST='0.0' GLINCORR='UNDEFINED' INSTRUME='ACS' LTV1='-2048.0' LTV2='-1.0' NUMCOLS='2070.0' NUMROWS='2046.0' OBSTYPE='INTERNAL' PCTECORR='OMIT' PHOTCORR='OMIT' REFTYPE='UNDEFINED' SHADCORR='OMIT' SHUTRPOS='B' TIME-OBS='01:07:14.960000' XCORNER='1.0' YCORNER='2072.0'
    0
    >>> config.set_crds_state(old_state)
    """

# ==================================================================================


def tst():
    """Run module tests,  for now just doctests only."""
    from crds.tests import test_matches, tstmod
    return tstmod(test_matches)

if __name__ == "__main__":
    print(tst())



"""
This module supports running crds.bestrefs from the command line using -m syntax:

% crds bestrefs ...

"""
import sys

from crds.bestrefs import bestrefs

def main():
    """Construct and run the bestrefs script,  return 1 if errors occurred, 0 otherwise."""
    errors = bestrefs.BestrefsScript()()
    exit_status = int(errors > 0)  # no errors = 0,  errors = 1
    return exit_status

if __name__ == "__main__":
    sys.exit(main())



"""This module defines rmap matching and generation customizations for WFPC2."""

from crds.core import log, rmap

'''
def _precondition_header_flatfile(header_in):
    """Mutate the incoming dataset header based upon hard coded rules
    and the header's contents.
    """
    header = dict(header_in)
    subarray = header["SUBARRAY"]
    aperture = header["APERTURE"]
    if subarray == "T" and "SUB" not in aperture:
        header["APERTURE"] = "*"
    return header

def precondition_header(rmap, header):
    if rmap.filekind == "flatfile":
        return _precondition_header_flatfile(header)
    else:
        return header
'''

def fallback_header_wfpc2_flatfile_v1(rmap, header):
    """Compute a fallback header for WFPC2 BIASFILE."""
    filter1 = header["FILTER1"]
    filter2 = header["FILTER2"]
    log.verbose("Computing fallback header wfpc2 ", rmap.filekind,
                "swapping filter1 was" , filter1, "filter2 was", filter2)
    header["FILTER1"] = filter2
    header["FILTER2"] = filter1
    return header

'''
         if lrfwave > 3000 and lrfwave <= 4200:
            filename = "M3C10045U.R4H"
         elif lrfwave > 4200 and lrfwave <= 5800:
            filename = "M3C1004FU.R4H"
         elif lrfwave > 5800 and lrfwave <= 7600:
            filename = "M3C1004NU.R4H"
         elif lrfwave > 7600 and lrfwave <= 10000:
            filename = "M3C10052U.R4H"
'''

def wfpc2_flatfile_filter(kmap):
    log.info("Hacking WFPC2 Flatfile.")
    # :  ('MODE', 'FILTER1', 'FILTER2', 'IMAGETYP', 'FILTNAM1', 'FILTNAM2', 'LRFWAVE'), ('DATE-OBS', 'TIME-OBS')),
    kmap[('*',    '*',       '*',       'EXT',       'FR*',     '*',      '# >3000 and <=4200 #')] = \
           [rmap.Filemap(date='1990-01-01 00:00:00', file='m3c10045u.r4h', comment='')]
    kmap[('*',    '*',       '*',       'EXT',       '*',     'FR*',      '# >3000 and <=4200 #')] = \
           [rmap.Filemap(date='1990-01-01 00:00:00', file='m3c10045u.r4h', comment='')]

    kmap[('*',    '*',       '*',       'EXT',       'FR*',     '*',      '# >4200 and <=5800 #')] = \
           [rmap.Filemap(date='1990-01-01 00:00:00', file='m3c1004fu.r4h', comment='')]
    kmap[('*',    '*',       '*',       'EXT',       '*',     'FR*',      '# >4200 and <=5800 #')] = \
           [rmap.Filemap(date='1990-01-01 00:00:00', file='m3c1004fu.r4h', comment='')]

    kmap[('*',    '*',       '*',       'EXT',       'FR*',     '*',      '# >5800 and <=7600 #')] = \
           [rmap.Filemap(date='1990-01-01 00:00:00', file='m3c1004nu.r4h', comment='')]
    kmap[('*',    '*',       '*',       'EXT',       '*',     'FR*',      '# >5800 and <=7600 #')] = \
           [rmap.Filemap(date='1990-01-01 00:00:00', file='m3c1004nu.r4h', comment='')]

    kmap[('*',    '*',       '*',       'EXT',       'FR*',     '*',      '# >7600 and <=10000 #')] = \
           [rmap.Filemap(date='1990-01-01 00:00:00', file='m3c10052u.r4h', comment='')]
    kmap[('*',    '*',       '*',       'EXT',       '*',     'FR*',      '# >7600 and <=10000 #')] = \
           [rmap.Filemap(date='1990-01-01 00:00:00', file='m3c10052u.r4h', comment='')]

    header_additions = [
        ("hooks", {
            "fallback_header" : "fallback_header_wfpc2_flatfile_v1",
        }),
    ]

    return kmap, header_additions


'''
      def wfpc2_flat_file_selection(self, querynum, thereffile, aSource):

    querytxt = ""
    #
    # add the file selection fields
    for k in thereffile._file_selections:
      if k._restrictions:
        # see if the restriction allows this file selection field
        opusutil.PrintMsg("D",'found a file select restricted: '+
                               k._restrictions)
        if (not eval(k._restrictions)):   # CDBS comment code
          opusutil.PrintMsg("D","File_selection on "+k._field+
                                " restricted away")
          continue
      #
      # skip these (not used in selection, only elsewhere for special tests)
      if (k._field == 'IMAGETYP' or
          k._field == 'FILTNAM1' or
          k._field == 'FILTNAM2' or
          k._field == 'LRFWAVE') :
          continue
      #
      # apply the file selection field
      # (first as a string, but if that fails, as a number converted to string,
      #  since numbers are not quoted but string are)
      try:
        querytxt = (querytxt + "and wfpc2_row_"+querynum+"." +
                 string.lower(k._field) + " = '" +
                 aSource._keywords[k._field][0] + "' ")
      except TypeError:
        querytxt = (querytxt + "and wfpc2_row_"+querynum+"." +
                 string.lower(k._field) + " = " +
                 str(aSource._keywords[k._field][0]) + " ")
    #
    return querytxt


  """
=======================================================================
Name: wfpc2_flatfile_custom

Description:
------------
This customized function composes a database query against the CDBS
database to find an applicable WFPC2 FLAT reference file.   A special
query is needed because the LRFWAVE keyword value can trigger a set of
hard-coded file selections, and if not, the two filter values can occur in
either order.

Arguments:
----------
thereffile (I) - a Reffile object containing information about the reference
                 file and its selection parameters
aSource (I)   - Input source object containing a dictionary of
                 keyword name/value pairs.  These form the
                 inputsource for the database query parameters.
Returns:
--------
reference_filename - name of the first reference file found by the DB query
                     (normally only one is found)

History:
--------
07/08/09 63053 MSwam     Initial version
=======================================================================
  """
  def wfpc2_flatfile_custom(self, thereffile, aSource):
    # the query templates
    # (no need to define these outside, like those for the generic case, since
    #  this routine is only called once per dataset)
    query_template_a = (
       "SELECT max(wfpc2_file_1.file_name) "+
       "FROM wfpc2_file wfpc2_file_1, wfpc2_row wfpc2_row_1 "+
       "WHERE wfpc2_file_1.useafter_date = "+
           "(SELECT max(wfpc2_file_2.useafter_date) "+
           "FROM wfpc2_file wfpc2_file_2, wfpc2_row wfpc2_row_2 "+
               "WHERE wfpc2_file_2.reference_file_type = 'FLT' "+
               "and wfpc2_file_2.useafter_date <= 'EXPOSURE_START' "+
               "and wfpc2_file_2.reject_flag = 'N' "+
               "and wfpc2_file_2.opus_flag = 'Y' "+
               "and wfpc2_file_2.archive_date is not null "+
               "and wfpc2_file_2.opus_load_date is not null "+
               "and wfpc2_file_2.file_name =wfpc2_row_2.file_name "+
               "and wfpc2_file_2.expansion_number = wfpc2_row_2.expansion_number ")
    #
    query_template_b = (
           ") and wfpc2_file_1.reference_file_type = 'FLT' "+
       "and wfpc2_file_1.reject_flag = 'N' "+
       "and wfpc2_file_1.opus_flag = 'Y' "+
       "and wfpc2_file_1.archive_date is not null "+
       "and wfpc2_file_1.opus_load_date is not null "+
       "and wfpc2_file_1.file_name =wfpc2_row_1.file_name "+
       "and wfpc2_file_1.expansion_number = wfpc2_row_1.expansion_number ")

    querytxt = query_template_a

    # fill the exposure start time, after converting keyword value
    # to datetime format from source format
    #
    exposure_start, dontcare = self.find_exposure_start(aSource)
    querytxt = string.replace(querytxt,"EXPOSURE_START", exposure_start)

    # add the file selection fields (row_2)
    querytxt = querytxt + self.wfpc2_flat_file_selection("2", thereffile,
                                                        aSource)
    # add second template
    querytxt = querytxt + query_template_b

    # add the file selection fields again (row_1)
    querytxt = querytxt + self.wfpc2_flat_file_selection("1", thereffile,
                                                        aSource)

    # replace any None values with null
    query1 = string.replace(querytxt, "None", "null")

    # get results in a list of lists
    result = [[]]
    self.zombie_select(query1, result)
    if len(result) == 0 or result[0][0] == None:
      #
      # no matching CDBS record found
      opusutil.PrintMsg("E","No matching FLAT found.")
      raise ZeroRowsFound, query1
    #
    # return the first filename found
    return result[0][0]

  """
=======================================================================
Name: wfpc2_flatfile

Description:
------------
This special-purpose function handles the WFPC2 FLATFILE.
In addition to handling the possibility that the WFPC2 filters can occur in
"normal" or "reversed" order, this routine also include special handling
for linear ramp filters (LRF), which get their flat selected based
on the LRFWAVE keyword.  In fact, because the LRFWAVE selection only
affects a small subset of data, and because it was added during the
final stages of WFPC-2 closeout processing, it was NOT added to
CDBS, and the names of the actual FLATs applied in these special cases
WILL BE HARD-CODED HERE.

The filter-ordering case is handled by making
a normal lookup first, and if that works then nothing else special
is done.  If it fails, then the order of the filter values is switched
and the query is tried again.  This is needed because the WFPC-2 filter order
is irrelevant for taking and collecting the science data, but is relevant in
the CDBS FLATs that have been delivered.  The delivered FLATS should have
supported the filters in either order, through an expansion, but that was
not done and there are too many of them now to reasonably expect the
entire set to be redelivered to CDBS.

Arguments:
----------
thereffile (I) - a Reffile object containing information about the reference
                 file and its selection parameters
aSource (I)   - Input source object containing a dictionary of
                 keyword name/value pairs.  These form the
                 inputsource for the database query parameters.
Returns:
--------
reference_filename - name of the first reference file found by the DB query
                     (normally only one is found)

History:
--------
03/37/03 46981 MSwam     another special case
07/08/09 63053 MSwam     add LRFWAVE handling
08/19/09 63351 MSwam     fix non-existing filter1 or filter2 case
=======================================================================
  """
  def wfpc2_flatfile(self, thereffile, aSource):
    try:
      # SPECIAL HARD-CODED LRFWAVE HANDLING CASE FIRST:
      #
      # If a linear ramp filter is being used, then the flat field
      # filename is HARD-CODED in the block below, and the normal
      # CDBS file selection algorithm is not used, since the LRFWAVE
      # value is not known to CDBS.
      #
      #print("filtnam1="+str(aSource._keywords['FILTNAM1']))
      #print("filtnam2="+str(aSource._keywords['FILTNAM2']))
      if (aSource._keywords['IMAGETYP'][0] == "EXT" and
          ((aSource._keywords['FILTNAM1'] != [None,0] and
            aSource._keywords['FILTNAM1'][0][0:2] == "FR") or
           (aSource._keywords['FILTNAM2'] != [None,0] and
            aSource._keywords['FILTNAM2'][0][0:2] == "FR") )):

         opusutil.PrintMsg("I","FLATFILE selection will be based on LRFWAVE value = "+str(aSource._keywords['LRFWAVE'][0]))
         lrfwave = float(aSource._keywords['LRFWAVE'][0])

         # THIS COULD be read from an external file, which would give
         # Ops the chance to override it, if it ever needs changes, but we
         # hear it will not be changed.
         #
         if lrfwave > 3000 and lrfwave <= 4200:
            filename = "M3C10045U.R4H"
         elif lrfwave > 4200 and lrfwave <= 5800:
            filename = "M3C1004FU.R4H"
         elif lrfwave > 5800 and lrfwave <= 7600:
            filename = "M3C1004NU.R4H"
         elif lrfwave > 7600 and lrfwave <= 10000:
            filename = "M3C10052U.R4H"
         else:
            opusutil.PrintMsg("E","FLATFILE selection FAILED for LRFWAVE value = "+aSource._keywords['LRFWAVE'][0])
            filename = " "
         return filename

      # reaching here means this is NOT a LRFWAVE case,
      #   so use the normal CDBS file selection method first
      #
      filename = self.wfpc2_flatfile_custom(thereffile, aSource)

    except ZeroRowsFound, querytxt:
      # normal query didn't work, try flipping the filters
      tmp = aSource._keywords['FILTER1'][0]
      aSource._keywords['FILTER1'][0] = aSource._keywords['FILTER2'][0]
      aSource._keywords['FILTER2'][0] = tmp
      opusutil.PrintMsg("D","FLATFILE not found for default filter order; now trying them reversed")
      filename = self.wfpc2_flatfile_custom(thereffile, aSource)
    #
    return filename
'''



"""Some simple console message functions w/counting features for
errors, warnings, and info.  Also error exception raising and
tracebacks.

>>> from crds.core import log
>>> log.set_test_mode()
>>> log.reset()

>>> log.warning("this is a test warning.")
CRDS - WARNING - this is a test warning.

>>> log.error("this is a test error.")
CRDS - ERROR - this is a test error.

>>> log.error("this is another test error.")
CRDS - ERROR - this is another test error.

>>> log.info("this is just informative.")
CRDS - INFO - this is just informative.

>>> log.errors()
2

>>> log.status()
(2, 1, 1)

>>> log.standard_status()
CRDS - INFO - 2 errors
CRDS - INFO - 1 warnings
CRDS - INFO - 1 infos

By default verbose messages are not emitted:

>>> log.verbose("this is a test verbose message.")

Calling set_verbose() turns on default verbosity=50:

>>> old_verbose = log.set_verbose()

>>> log.verbose("this is a test verbose message.")
CRDS - DEBUG - this is a test verbose message.

No output is expected since default verbosity=50:

>>> log.verbose("this is a test suppressed verbose 60 message.", verbosity=60)

Output should now occur since verbosity=60:

>>> log.set_verbose(60)
50
>>> log.verbose("this is a test verbose 60 message.", verbosity=60)
CRDS - DEBUG - this is a test verbose 60 message.

A number of context managers are defined for succinctly mapping nested
exceptions onto CRDS messages or adding information:

>>> with log.error_on_exception("Something bad happened and we trapped it"):
...     raise ValueError("some value was bad.")
CRDS - ERROR - Something bad happened and we trapped it : some value was bad.

>>> with log.augment_exception("A tad more parent info"):
...     raise ValueError("some vague deeply nested exception.")
Traceback (most recent call last):
...
ValueError: A tad more parent info : some vague deeply nested exception.

When the exception trap mode is set to False,  the exception context managers
let the exceptions pass through unmodified:

>>> old_trap = log.set_exception_trap(False)

>>> with log.error_on_exception("Something bad happened and we trapped it"):
...     raise ValueError("some value was bad.")
Traceback (most recent call last):
...
ValueError: some value was bad.

>>> with log.error_on_exception("Something bad happened and we trapped it"):
...     raise ValueError("some value was bad.")
...     print("this code not reached")
Traceback (most recent call last):
...
ValueError: some value was bad.

>>> with log.augment_exception("A tad more parent info"):
...     raise ValueError("some vague deeply nested exception.")
Traceback (most recent call last):
...
ValueError: some vague deeply nested exception.

Command line tools control the exception trap flag using --debug-traps,  this is
useful for debugging anomalies during type (or tool) development.   The key points
about the exception trap flag are that (a) traps are passed through and (b) exceptions
are carefully reraised so the original traceback is not destroyed.

>>> _ = log.set_verbose(old_verbose)
>>> _ = log.set_exception_trap(old_trap)

>>> old_verbose = log.set_verbose()

>>> with log.verbose_warning_on_exception("Testing expected verbose warning"):
...     raise ValueError("Force verbose warning.")
...     log.write("This code should not be exected.")
CRDS - WARNING -  Testing expected verbose warning : Force verbose warning.

>>> _ = log.set_verbose(old_verbose)
"""
import sys
import os
import optparse
import logging
import pprint
import contextlib

DEFAULT_VERBOSITY_LEVEL = 50

class CrdsLogger:
    def __init__(self, name="CRDS", enable_console=True, level=logging.DEBUG, enable_time=True):
        self.name = name

        self.handlers = []  # logging handlers, used e.g. to add console or file output streams
        self.filters = []   # simple CRDS filters, used e.g. to mutate message text

        self.logger = logging.getLogger(name)
        self.logger.setLevel(level)
        self.logger.propagate = False
        self.formatter = self.set_formatter()
        self.console = None
        if enable_console:
            self.add_console_handler(level)

        # CRDS internal counters of types of messages
        self.errors = 0
        self.warnings = 0
        self.infos = 0
        self.debugs = 0

        self.eol_pending = False

        # verbose_level handles CRDS verbosity,  defaulting to 0 for no debug
        try:
            verbose_level = os.environ.get("CRDS_VERBOSITY", 0)
            self.verbose_level =  int(verbose_level)
        except Exception:
            warning("Bad format for CRDS_VERBOSITY =", repr(verbose_level),
                        "Use e.g. -1 to squelch info, 0 for no debug,  50 for default debug output. 100 max debug.")
            self.verbose_level = DEFAULT_VERBOSITY_LEVEL

    def set_formatter(self, enable_time=False, enable_msg_count=True):
        """Set the formatter attribute of `self` to a logging.Formatter and return it."""
        self.formatter = logging.Formatter(
            '{}%(name)s - %(levelname)s - %(message)s'.format(
                "%(asctime)s - " if enable_time else "",
                ))
        for handler in self.handlers:
            handler.setFormatter(self.formatter)
        return self.formatter

    @property
    def msg_count(self):
        return "(%07d) -" % (self.infos + self.errors + self.warnings + self.debugs) if ADD_LOG_MSG_COUNT else ""

    def format(self, *args, **keys):
        end = keys.get("end", "\n")
        sep = keys.get("sep", " ")
        output = sep.join([str(arg) for arg in args]) + end
        for filter in self.filters:
            output = filter(output)
        return output

    def eformat(self, *args, **keys):
        keys["end"] = ""
        if self.eol_pending:
            self.write()
        return self.format(*args, **keys)

    def info(self, *args, **keys):
        self.infos += 1
        if self.verbose_level > -1:
            self.logger.info(self.eformat(self.msg_count, *args, **keys))

    def warn(self, *args, **keys):
        self.warnings += 1
        if self.verbose_level > -2:
            self.logger.warning(self.eformat(self.msg_count, *args, **keys))

    def error(self, *args, **keys):
        self.errors += 1
        if self.verbose_level > -3:
            self.logger.error(self.eformat(self.msg_count, *args, **keys))

    def debug(self, *args, **keys):
        self.debugs += 1
        self.logger.debug(self.eformat(self.msg_count, *args, **keys))

    def should_output(self, *args, **keys):
        verbosity = keys.get("verbosity", DEFAULT_VERBOSITY_LEVEL)
        return not self.verbose_level < verbosity

    def verbose(self, *args, **keys):
        if self.should_output(*args, **keys):
            self.debug(*args, **keys)

    def verbose_warning(self, *args, **keys):
        if self.should_output(*args, **keys):
            self.warn(*args, **keys)

    def write(self, *args, **keys):
        """Output a message to stdout, formatting each positional parameter
        as a string.
        """
        output = self.format(*args, **keys)
        self.eol_pending = not output.endswith("\n")
        sys.stderr.flush()
        sys.stdout.write(output)
        sys.stdout.flush()

    def status(self):
        return self.errors, self.warnings, self.infos

    def reset(self):
        self.errors = self.warnings = self.infos = self.debugs = 0

    def set_verbose(self, level=True):
        assert -3 <= level <= 100,  "verbosity level must be in range -3..100"
        old_verbose = self.verbose_level
        if level == True:
            level = DEFAULT_VERBOSITY_LEVEL
        elif level == False:
            level = 0
        self.verbose_level = level
        return old_verbose

    def get_verbose(self):
        return self.verbose_level

    def add_console_handler(self, level=logging.DEBUG, stream=sys.stderr):
        if self.console is None:
            self.console = self.add_stream_handler(stream)

    def remove_console_handler(self):
        if self.console is not None:
            self.remove_stream_handler(self.console)
            self.console = None

    def add_stream_handler(self, filelike, level=logging.DEBUG):
        handler = logging.StreamHandler(filelike)
        handler.setLevel(level)
        handler.propagate = False
        handler.setFormatter(self.formatter)
        self.handlers.append(handler)
        self.logger.addHandler(handler)
        return handler

    def remove_stream_handler(self, handler):
        self.handlers.remove(handler)
        self.logger.removeHandler(handler)

    def fatal_error(self, *args, **keys):
        error("(FATAL)", *args, **keys)
        sys.exit(-1)  # FATAL == totally unambiguous

    def prepend_crds_filter(self, filter):
        """Prepend a message `filter` to CRDS log output independent of logging framework."""
        if filter not in self.filters:
            self.filters = [filter] + self.filters

    def append_crds_filter(self, filter):
        """Append a message `filter` to CRDS log output independent of logging framework."""
        if filter not in self.filters:
            self.filters = self.filters + [filter]

    def remove_crds_filter(self, filter):
        """Remove `filter` from the sequence of CRDS message filters."""
        if filter in self.filters:
            self.filters.remove(filter)

THE_LOGGER = CrdsLogger("CRDS")

info = THE_LOGGER.info
error = THE_LOGGER.error
warning = THE_LOGGER.warn
verbose_warning = THE_LOGGER.verbose_warning
verbose = THE_LOGGER.verbose
should_output = THE_LOGGER.should_output
debug = THE_LOGGER.debug
fatal_error = THE_LOGGER.fatal_error
status = THE_LOGGER.status
reset = THE_LOGGER.reset
write = THE_LOGGER.write
set_verbose = THE_LOGGER.set_verbose
get_verbose = THE_LOGGER.get_verbose
add_console_handler = THE_LOGGER.add_console_handler
remove_console_handler = THE_LOGGER.remove_console_handler
add_stream_handler = THE_LOGGER.add_stream_handler
remove_stream_handler = THE_LOGGER.remove_stream_handler
append_crds_filter = THE_LOGGER.append_crds_filter
remove_crds_filter = THE_LOGGER.remove_crds_filter
prepend_crds_filter = THE_LOGGER.prepend_crds_filter
format = THE_LOGGER.format

def increment_errors(N=1):
    """Increment the error count by N without issuing a log message."""
    THE_LOGGER.errors += N

def errors():
    """Return the global count of errors."""
    return THE_LOGGER.errors

def warnings():
    """Return the global count of errors."""
    return THE_LOGGER.warnings

def infos():
    """Return the global count of infos."""
    return THE_LOGGER.infos

def set_test_mode():
    """Route log messages to standard output for testing with doctest."""
    remove_console_handler()
    add_console_handler(stream=sys.stdout)
    set_log_time(False)

def set_log_time(enable_time=False):
    """Set the flag for including time in log messages.  Ignore CRDS_LOG_TIME."""
    THE_LOGGER.set_formatter(enable_time)

# ===========================================================================

@contextlib.contextmanager
def reduced_verbosity(reduced_level, threshhold):
    """Sets the global verbosity level to `reduced_level` as long as it is already below `threshhold`
    within the scope of the with-block / context manager.

    If the global verbosity is above `threshhold`,  don't set to `reduced_level` enabling high verbosity settings
    to bypass the suppression.
    """
    old_verbose = get_verbose()
    if old_verbose < threshhold:
        set_verbose(min(old_verbose, reduced_level))
    try:
        yield
    finally:
        if old_verbose < threshhold:
            set_verbose(old_verbose)

# ===========================================================================

def exception_trap_logger(func):
    @contextlib.contextmanager
    def func_on_exception(*args, **keys):
        """func_on_exception is a context manager which issues a func() message if any statement
        in a with-block generates an exception.   The exception is suppressed.

        >> with warn_on_exception("As expected, it failed."):
        ...    print("do it.")
        do it.

        >> with warn_on_exception("As expected, it failed."):
        ...    raise Exception("It failed!")
        ...    print("do it.")

        Never printed 'do it.'  Nothing printed because the func() output is a log message.
        """
        try:
            yield
        except Exception as exc:
            reraise = func(*args + (":", str(exc)), **keys)
            if not CRDS_EXCEPTION_TRAP:
                # In python-2, distinction between raise and "raise something".  raise doesn't
                # wreck the traceback,  raising a new improved exception does.
                raise
            # Augmented,  the traceback is trashed from here down but the message is better when caught higher up.
            elif reraise:
                exc_class = keys.pop("exception_class", exc.__class__)
                keys["end"] = ""
                raise exc_class(format(*args + (":", str(exc)), **keys)) from exc
            else:
                pass # snuff the exception,  func() probably issued a log message.
    return func_on_exception

CRDS_EXCEPTION_TRAP = True

def get_exception_trap():
    """Return the value of the exception trap flag,  False means 'debug'."""
    return CRDS_EXCEPTION_TRAP

def set_exception_trap(flag):
    """Set the debug by-pass mode flag for log.error_on_exception() exception trap."""
    global CRDS_EXCEPTION_TRAP
    old_flag = CRDS_EXCEPTION_TRAP
    if flag is not None:
        CRDS_EXCEPTION_TRAP = flag
    return old_flag

def _reraise(*args, **keys):
    """Signal to exception_trap_logger to unconditionally reraise the exception,  probably augmented."""
    return True

info_on_exception = exception_trap_logger(info)
debug_on_exception = exception_trap_logger(debug)
verbose_on_exception = exception_trap_logger(verbose)
verbose_warning_on_exception = exception_trap_logger(verbose_warning)
warn_on_exception = exception_trap_logger(warning)
error_on_exception = exception_trap_logger(error)
fatal_error_on_exception = exception_trap_logger(fatal_error)

# Either supress-with-error or augment the exception in the enclosed block.
# Suppressing an exception results in less information to identify the problem
# but can also reveal multiple problems in a single program run if executed
# inside a loop that an augmented exception would have exited.
if bool(os.environ.get("CRDS_TRAP_AUGMENTED", False)):
    augment_exception = error_on_exception
else:
    augment_exception = exception_trap_logger(_reraise)

# ===========================================================================

ADD_LOG_MSG_COUNT = False

def set_add_log_msg_count(flag):
    global ADD_LOG_MSG_COUNT
    old_flag = ADD_LOG_MSG_COUNT
    ADD_LOG_MSG_COUNT = flag
    return old_flag

def get_add_log_msg_count():
    return ADD_LOG_MSG_COUNT

# ===========================================================================

class PP:
    """A wrapper to defer pretty printing until after it's known a verbose
    message will definitely be output.
    """
    def __init__(self, ppobj):
        self.ppobj = ppobj

    def __str__(self):
        return pprint.pformat(self.ppobj)

class Deferred:
    """A wrapper to delay calling a callable until after it's known a verbose
    message will definitely be output.
    """
    def __init__(self, ppobj):
        self.ppobj = ppobj

    def __str__(self):
        return str(self.ppobj())

# ===========================================================================

def handle_standard_options(
        args, parser=None, usage="usage: %prog [options] <inpaths...>"):
    '''Set some standard options on an optparser object,  many
    of which interplay with the log module itself.
    '''
    if parser is None:
        parser = optparse.OptionParser(usage)

    parser.add_option("-V", "--verbose", dest="verbose",
                      help="Set verbosity level.",
                      metavar="VERBOSITY", default=None)

    options, args = parser.parse_args(args)

    if options.verbose is not None:
        set_verbose(int(options.verbose))

    return options, args

def standard_status():
    """Print out errors, warnings, and infos."""
    errors, warnings, infos = THE_LOGGER.status()
    info(errors, "errors")
    info(warnings, "warnings")
    info(infos, "infos")

# ==============================================================================

def format_parameter_list(parameters):
    """Given an item list or dictionary of matching parameters,  return a formatted string.
    Typically used to format matching parameters or bestrefs dicts for debug.
    """
    items = sorted(dict(parameters).items())
    return " ".join(["=".join([key, repr(str(value))]) for (key,value) in items])

# ==============================================================================

def srepr(obj):
    """Return the repr() of the str() of obj"""
    return repr(str(obj))

# ==============================================================================

def divider(name="", char="-", n=75, func=info, **keys):
    """Create a log divider line consisting of `char` repeated `n` times
    possibly with `name` injected into the center of the divider.
    Output it as a string to logging function `func` defaulting to info().
    """
    if name:
        n2 = (n - len(name) - 2) // 2
        func(char*n2, name, char*n2, **keys)
    else:
        func(char*n, **keys)

# ===================================================================

def test():
    from crds.core import log
    import doctest
    return doctest.testmod(log)

if __name__ == "__main__":
    print(test())



import doctest

from crds.core import log
from crds.client import api
from crds.misc.query_affected import QueryAffectedDatasetsScript

from crds.tests import test_config

def dt_get_context_history():
    """
    Test the web service which gets the context history from the JWST server for
    ticket #360, "Verify CRDS correctly reports historical configurations".

    Note this test is dependent on the JWST operational server.   A similar test
    can be performed for any of the I&T servers by changing the URL.   In I&T,
    setup for this test will involve correctly configuring CRDS_PATH and
    CRDS_SERVER_URL to point to the I&T CRDS cache and I&T CRDS Server respectively.

    >>> old_state = test_config.setup(url='https://jwst-crds.stsci.edu')
    >>> history = api.get_context_history("jwst")

    >>> history[0]
    ('2012-09-06 00:00:00', 'jwst.pmap', 'Bootstrap mappings')

    >>> history[32]
    ('2015-11-18 12:58:13', 'jwst_0105.pmap', 'Declared various EXP_TYPE as N/A for 15 WCS types for MIRI, NIRCAM, NIRSPEC. Replacement MIRI distortion references for CRDS #238.')

    >>> test_config.cleanup(old_state)
    """

def dt_query_affected_datasets_list():
    """
    This example demos the "--list" feature of the query_affected_datasets script used
    to query the CRDS reprocessing system for datasets affected by a context transition.
    Since the context history is integral to the affected datasets computation,  the query
    tool has a mode for listing the context history to orient any human operator.

    >>> old_state = test_config.setup(url='https://jwst-crds.stsci.edu')

    >>> QueryAffectedDatasetsScript("query_affected --list -x 0 -y 10")()
    (0, '2012-09-06 00:00:00', 'jwst.pmap', 'Bootstrap mappings')
    (1, '2012-09-27 00:00:00', 'jwst_0000.pmap', 'First rules and references from jwst_gentools stub development cloning.')
    (2, '2013-04-13 00:00:00', 'jwst_0001.pmap', 'Linearity and dark files.')
    (3, '2013-07-31 00:00:00', 'jwst_0002.pmap', 'Dark and Mask files.')
    (4, '2013-09-04 00:00:00', 'jwst_0003.pmap', 'Absolute Calibration (photom) additions and replacements.')
    (5, '2013-11-25 09:00:03', 'jwst_0005.pmap', 'set by system')
    (6, '2014-03-19 10:51:19', 'jwst_0012.pmap', 'Updated for META.INSTRUMENT.TYPE switch to META.INSTRUMENT.NAME\\r\\nNew linearity files for all instruments\\r\\nNew saturation files and rmaps for all instruments')
    (7, '2014-03-26 08:49:12', 'jwst_0013.pmap', 'New context jwst_0013.pmap adds Interpixel Capacitance references for all instruments.')
    (8, '2014-07-08 16:09:39', 'jwst_0022.pmap', 'Multiple MIRI FLAT changes, new references, added SUBARRAY as matching parameter.  See diffs.')
    (9, '2014-07-18 11:34:48', 'jwst_0023.pmap', 'First MIRI PHOTOM reference for MIRIMAGE detector with real data.')
    (10, '2014-07-24 18:49:03', 'jwst_0025.pmap', 'Replaced PHOTOM references for NIRSPEC and NIRISS.')

    >>> test_config.cleanup(old_state)
    """



def main():
    """Run module tests,  for now just doctests only.

    test_config.setup() and cleanup() are done inline above because bracketing
    the tests here does not get picked up by nose test discovery.  Combining
    tests into one giant docstring works but is hard to analyze and debug when
    things go wrong.
    """
    from crds.tests import test_build6, tstmod
    return tstmod(test_build6)

if __name__ == "__main__":
    print(main())
